apiset_win10 = {
    'api-ms-onecoreuap-print-render-l1-1-0': {None: 'printrenderapihost.dll'},
    'api-ms-onecoreuap-settingsync-status-l1-1-0': {None: 'settingsynccore.dll'},
    'api-ms-win-appmodel-identity-l1-2-0': {None: 'kernel.appcore.dll'},
    'api-ms-win-appmodel-runtime-internal-l1-1-4': {None: 'kernel.appcore.dll'},
    'api-ms-win-appmodel-runtime-l1-1-2': {None: 'kernel.appcore.dll'},
    'api-ms-win-appmodel-state-l1-1-2': {None: 'kernel.appcore.dll'},
    'api-ms-win-appmodel-state-l1-2-0': {None: 'kernel.appcore.dll'},
    'api-ms-win-appmodel-unlock-l1-1-0': {None: 'kernel.appcore.dll'},
    'api-ms-win-base-bootconfig-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-base-util-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-composition-redirection-l1-1-0': {None: 'dwmredir.dll'},
    'api-ms-win-composition-windowmanager-l1-1-0': {None: 'udwm.dll'},
    'api-ms-win-core-apiquery-l1-1-0': {None: 'ntdll.dll'},
    'api-ms-win-core-appcompat-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-appinit-l1-1-0': {None: 'kernel32.dll',
                                       'kernel32.dll': 'kernelbase.dll'},
    'api-ms-win-core-atoms-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-bicltapi-l1-1-4': {None: 'bi.dll'},
    'api-ms-win-core-biplmapi-l1-1-5': {None: 'twinapi.appcore.dll'},
    'api-ms-win-core-biptcltapi-l1-1-5': {None: 'twinapi.appcore.dll'},
    'api-ms-win-core-calendar-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-com-l1-1-2': {None: 'combase.dll'},
    'api-ms-win-core-com-l2-1-1': {None: 'coml2.dll'},
    'api-ms-win-core-com-midlproxystub-l1-1-0': {None: 'combase.dll'},
    'api-ms-win-core-com-private-l1-1-1': {None: 'combase.dll'},
    'api-ms-win-core-comm-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-console-ansi-l2-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-console-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-console-l2-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-console-l3-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-crt-l1-1-0': {None: 'ntdll.dll'},
    'api-ms-win-core-crt-l2-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-datetime-l1-1-2': {None: 'kernelbase.dll'},
    'api-ms-win-core-debug-l1-1-2': {None: 'kernelbase.dll'},
    'api-ms-win-core-debug-minidump-l1-1-0': {None: 'dbgcore.dll'},
    'api-ms-win-core-delayload-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-enclave-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-errorhandling-l1-1-3': {None: 'kernelbase.dll'},
    'api-ms-win-core-featurestaging-l1-1-1': {None: 'shcore.dll'},
    'api-ms-win-core-fibers-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-fibers-l2-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-file-ansi-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-file-ansi-l2-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-file-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-file-l1-2-2': {None: 'kernelbase.dll'},
    'api-ms-win-core-file-l2-1-2': {None: 'kernelbase.dll'},
    'api-ms-win-core-firmware-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-guard-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-handle-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-heap-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-heap-l1-2-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-heap-l2-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-heap-obsolete-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-interlocked-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-interlocked-l1-2-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-io-l1-1-1': {None: 'kernel32.dll',
                                  'kernel32.dll': 'kernelbase.dll'},
    'api-ms-win-core-job-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-job-l2-1-1': {None: 'kernel32.dll'},
    'api-ms-win-core-kernel32-legacy-ansi-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-kernel32-legacy-l1-1-5': {None: 'kernel32.dll'},
    'api-ms-win-core-kernel32-private-l1-1-2': {None: 'kernel32.dll'},
    'api-ms-win-core-largeinteger-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-libraryloader-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-libraryloader-l1-2-2': {None: 'kernelbase.dll'},
    'api-ms-win-core-libraryloader-l2-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-libraryloader-private-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-localization-ansi-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-localization-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-localization-l1-2-2': {None: 'kernelbase.dll'},
    'api-ms-win-core-localization-l2-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-localization-obsolete-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-localization-obsolete-l1-2-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-localization-obsolete-l1-3-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-localization-private-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-localregistry-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-marshal-l1-1-0': {None: 'combase.dll'},
    'api-ms-win-core-memory-l1-1-5': {None: 'kernelbase.dll'},
    'api-ms-win-core-misc-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-multipleproviderrouter-l1-1-0': {None: 'mpr.dll'},
    'api-ms-win-core-namedpipe-ansi-l1-1-1': {None: 'kernel32.dll'},
    'api-ms-win-core-namedpipe-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-namedpipe-l1-2-2': {None: 'kernelbase.dll'},
    'api-ms-win-core-namespace-ansi-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-namespace-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-normalization-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-path-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-perfcounters-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-privateprofile-l1-1-1': {None: 'kernel32.dll'},
    'api-ms-win-core-processenvironment-ansi-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-processenvironment-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-processenvironment-l1-2-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-processsecurity-l1-1-0': {None: 'kernel32.dll',
                                               'kernel32.dll': 'kernelbase.dll'},
    'api-ms-win-core-processsnapshot-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-processthreads-l1-1-3': {None: 'kernel32.dll',
                                              'kernel32.dll': 'kernelbase.dll'},
    'api-ms-win-core-processtopology-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-processtopology-l1-2-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-processtopology-obsolete-l1-1-1': {None: 'kernel32.dll'},
    'api-ms-win-core-processtopology-private-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-profile-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-psapi-ansi-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-psapi-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-psapi-obsolete-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-psapiansi-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-psm-app-l1-1-0': {None: 'twinapi.appcore.dll'},
    'api-ms-win-core-psm-appnotify-l1-1-0': {None: 'twinapi.appcore.dll'},
    'api-ms-win-core-psm-info-l1-1-1': {None: 'appsruprov.dll'},
    'api-ms-win-core-psm-key-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-psm-plm-l1-1-3': {None: 'twinapi.appcore.dll'},
    'api-ms-win-core-psm-plm-l1-2-0': {None: 'twinapi.appcore.dll'},
    'api-ms-win-core-psm-rtimer-l1-1-1': {None: 'twinapi.appcore.dll'},
    'api-ms-win-core-psm-tc-l1-1-1': {None: 'twinapi.appcore.dll'},
    'api-ms-win-core-quirks-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-realtime-l1-1-2': {None: 'kernelbase.dll'},
    'api-ms-win-core-registry-fromapp-l1-1-0': {None: 'reguwpapi.dll'},
    'api-ms-win-core-registry-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-registry-l2-1-0': {None: 'advapi32.dll'},
    'api-ms-win-core-registry-l2-2-0': {None: 'advapi32.dll'},
    'api-ms-win-core-registry-private-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-core-registryuserspecific-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-rtlsupport-l1-1-1': {None: 'ntdll.dll'},
    'api-ms-win-core-rtlsupport-l1-2-0': {None: 'ntdll.dll'},
    'api-ms-win-core-shlwapi-legacy-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-shlwapi-obsolete-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-shlwapi-obsolete-l1-2-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-shutdown-ansi-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-core-shutdown-l1-1-1': {None: 'advapi32.dll'},
    'api-ms-win-core-sidebyside-ansi-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-sidebyside-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-slapi-l1-1-0': {None: 'clipc.dll'},
    'api-ms-win-core-string-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-string-l2-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-string-obsolete-l1-1-1': {None: 'kernel32.dll'},
    'api-ms-win-core-stringansi-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-stringloader-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-synch-ansi-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-synch-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-synch-l1-2-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-sysinfo-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-sysinfo-l1-2-3': {None: 'kernelbase.dll'},
    'api-ms-win-core-sysinfo-l2-1-0': {None: 'advapi32.dll'},
    'api-ms-win-core-systemtopology-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-textinput-client-l1-1-0': {None: 'textinputframework.dll'},
    'api-ms-win-core-threadpool-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-threadpool-l1-2-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-threadpool-legacy-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-threadpool-private-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-timezone-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-timezone-private-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-toolhelp-l1-1-1': {None: 'kernel32.dll'},
    'api-ms-win-core-ums-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-core-url-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-util-l1-1-1': {None: 'kernel32.dll',
                                    'kernel32.dll': 'kernelbase.dll'},
    'api-ms-win-core-version-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-version-private-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-versionansi-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-windowsceip-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-core-windowserrorreporting-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-winrt-error-l1-1-1': {None: 'combase.dll'},
    'api-ms-win-core-winrt-errorprivate-l1-1-1': {None: 'combase.dll'},
    'api-ms-win-core-winrt-l1-1-0': {None: 'combase.dll'},
    'api-ms-win-core-winrt-propertysetprivate-l1-1-1': {None: 'wintypes.dll'},
    'api-ms-win-core-winrt-registration-l1-1-0': {None: 'combase.dll'},
    'api-ms-win-core-winrt-robuffer-l1-1-0': {None: 'wintypes.dll'},
    'api-ms-win-core-winrt-roparameterizediid-l1-1-0': {None: 'combase.dll'},
    'api-ms-win-core-winrt-string-l1-1-1': {None: 'combase.dll'},
    'api-ms-win-core-wow64-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-core-xstate-l1-1-2': {None: 'ntdll.dll'},
    'api-ms-win-core-xstate-l2-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-coremessaging-host-l1-1-0': {None: ''},
    'api-ms-win-coreui-secruntime-l1-1-0': {None: ''},
    'api-ms-win-crt-conio-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-convert-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-environment-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-filesystem-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-heap-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-locale-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-math-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-multibyte-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-private-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-process-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-runtime-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-stdio-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-string-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-time-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-crt-utility-l1-1-0': {None: 'ucrtbase.dll'},
    'api-ms-win-deprecated-apis-advapi-l1-1-0': {None: ''},
    'api-ms-win-deprecated-apis-legacy-l1-1-0': {None: ''},
    'api-ms-win-deprecated-apis-legacy-l1-2-0': {None: ''},
    'api-ms-win-deprecated-apis-obsolete-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-devices-config-l1-1-1': {None: 'cfgmgr32.dll'},
    'api-ms-win-devices-query-l1-1-1': {None: 'cfgmgr32.dll'},
    'api-ms-win-devices-swdevice-l1-1-1': {None: 'cfgmgr32.dll'},
    'api-ms-win-downlevel-advapi32-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-downlevel-advapi32-l2-1-1': {None: 'sechost.dll'},
    'api-ms-win-downlevel-advapi32-l3-1-0': {None: 'ntmarta.dll'},
    'api-ms-win-downlevel-advapi32-l4-1-0': {None: 'advapi32.dll'},
    'api-ms-win-downlevel-kernel32-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-downlevel-kernel32-l2-1-0': {None: 'kernel32.dll'},
    'api-ms-win-downlevel-normaliz-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-downlevel-ole32-l1-1-1': {None: 'combase.dll'},
    'api-ms-win-downlevel-shell32-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-downlevel-shlwapi-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-downlevel-shlwapi-l2-1-1': {None: 'shcore.dll'},
    'api-ms-win-downlevel-user32-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-downlevel-version-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-dwmapi-l1-1-0': {None: 'dwmapi.dll'},
    'api-ms-win-dx-d3dkmt-l1-1-3': {None: 'gdi32.dll'},
    'api-ms-win-eventing-classicprovider-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-eventing-consumer-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-eventing-controller-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-eventing-legacy-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-eventing-obsolete-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-eventing-provider-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-eventing-tdh-l1-1-0': {None: 'tdh.dll', 'tdh.dll': 'mintdh.dll'},
    'api-ms-win-eventlog-legacy-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-eventlog-private-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-gaming-expandedresources-l1-1-0': {None: 'gamemode.dll'},
    'api-ms-win-gaming-gamemonitor-l1-1-0': {None: 'gamemonitor.dll'},
    'api-ms-win-gaming-tcui-l1-1-3': {None: 'gamingtcui.dll'},
    'api-ms-win-gdi-dpiinfo-l1-1-0': {None: 'gdi32.dll'},
    'api-ms-win-gdi-internal-uap-l1-1-0': {None: 'gdi32full.dll'},
    'api-ms-win-http-time-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-input-ie-interactioncontext-l1-1-0': {None: ''},
    'api-ms-win-legacy-shlwapi-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-mm-joystick-l1-1-0': {None: 'winmmbase.dll'},
    'api-ms-win-mm-mci-l1-1-0': {None: 'winmm.dll'},
    'api-ms-win-mm-misc-l1-1-1': {None: 'winmmbase.dll'},
    'api-ms-win-mm-misc-l2-1-0': {None: 'winmm.dll'},
    'api-ms-win-mm-mme-l1-1-0': {None: 'winmmbase.dll'},
    'api-ms-win-mm-playsound-l1-1-0': {None: 'winmm.dll'},
    'api-ms-win-mm-time-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-net-isolation-l1-1-1': {None: 'firewallapi.dll'},
    'api-ms-win-networking-interfacecontexts-l1-1-0': {None: 'ondemandconnroutehelper.dll'},
    'api-ms-win-ngc-serialization-l1-1-0': {None: 'ngckeyenum.dll'},
    'api-ms-win-ntuser-ie-message-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-ntuser-ie-window-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-ntuser-ie-wmpointer-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-ntuser-rectangle-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-ntuser-sysparams-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-obsolete-localization-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-obsolete-psapi-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-obsolete-shlwapi-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-ole32-ie-l1-1-0': {None: 'ole32.dll'},
    'api-ms-win-oobe-notification-l1-1-0': {None: 'kernel32.dll'},
    'api-ms-win-perf-legacy-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-power-base-l1-1-0': {None: 'powrprof.dll'},
    'api-ms-win-power-limitsmanagement-l1-1-0': {None: 'powrprof.dll'},
    'api-ms-win-power-setting-l1-1-0': {None: 'powrprof.dll'},
    'api-ms-win-ro-typeresolution-l1-1-0': {None: 'wintypes.dll'},
    'api-ms-win-rtcore-minuser-private-l1-1-1': {None: ''},
    'api-ms-win-rtcore-navigation-l1-1-0': {None: ''},
    'api-ms-win-rtcore-ntuser-clipboard-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-rtcore-ntuser-draw-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-rtcore-ntuser-powermanagement-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-rtcore-ntuser-private-l1-1-5': {None: 'user32.dll'},
    'api-ms-win-rtcore-ntuser-shell-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-rtcore-ntuser-synch-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-rtcore-ntuser-window-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-rtcore-ntuser-winevent-l1-1-0': {None: 'user32.dll'},
    'api-ms-win-rtcore-ntuser-wmpointer-l1-1-3': {None: 'user32.dll'},
    'api-ms-win-rtcore-ole32-clipboard-l1-1-1': {None: 'ole32.dll'},
    'api-ms-win-rtcore-session-l1-1-1': {None: ''},
    'api-ms-win-rtcore-session-l1-2-0': {None: ''},
    'api-ms-win-security-accesshlpr-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-security-activedirectoryclient-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-security-appcontainer-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-security-audit-l1-1-1': {None: 'sechost.dll'},
    'api-ms-win-security-base-ansi-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-security-base-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-security-base-l1-2-2': {None: 'kernelbase.dll'},
    'api-ms-win-security-base-private-l1-1-1': {None: 'kernelbase.dll'},
    'api-ms-win-security-capability-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-security-cpwl-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-security-credentials-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-security-credentials-l2-1-0': {None: 'sechost.dll'},
    'api-ms-win-security-cryptoapi-l1-1-0': {None: 'cryptsp.dll'},
    'api-ms-win-security-grouppolicy-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-security-isolatedcontainer-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-security-logon-l1-1-1': {None: 'advapi32.dll'},
    'api-ms-win-security-lsalookup-ansi-l2-1-0': {None: 'advapi32.dll'},
    'api-ms-win-security-lsalookup-l1-1-2': {None: 'sechost.dll'},
    'api-ms-win-security-lsalookup-l2-1-1': {None: 'advapi32.dll'},
    'api-ms-win-security-lsapolicy-l1-1-1': {None: 'sechost.dll'},
    'api-ms-win-security-provider-ansi-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-security-provider-l1-1-0': {None: 'ntmarta.dll'},
    'api-ms-win-security-sddl-ansi-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-security-sddl-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-security-sddlparsecond-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-security-systemfunctions-l1-1-0': {None: 'advapi32.dll'},
    'api-ms-win-security-trustee-l1-1-2': {None: 'advapi32.dll'},
    'api-ms-win-service-core-ansi-l1-1-1': {None: 'advapi32.dll'},
    'api-ms-win-service-core-l1-1-2': {None: 'sechost.dll'},
    'api-ms-win-service-management-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-service-management-l2-1-0': {None: 'sechost.dll'},
    'api-ms-win-service-private-l1-1-3': {None: 'sechost.dll'},
    'api-ms-win-service-winsvc-l1-1-0': {None: 'sechost.dll'},
    'api-ms-win-service-winsvc-l1-2-0': {None: 'sechost.dll'},
    'api-ms-win-shcore-comhelpers-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shcore-obsolete-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shcore-path-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shcore-registry-l1-1-1': {None: 'shcore.dll'},
    'api-ms-win-shcore-scaling-l1-1-2': {None: 'shcore.dll'},
    'api-ms-win-shcore-stream-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shcore-stream-winrt-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shcore-sysinfo-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shcore-thread-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shcore-unicodeansi-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shell-associations-l1-1-0': {None: 'windows.storage.dll'},
    'api-ms-win-shell-changenotify-l1-1-0': {None: 'windows.storage.dll'},
    'api-ms-win-shell-dataobject-l1-1-0': {None: 'windows.storage.dll'},
    'api-ms-win-shell-namespace-l1-1-0': {None: 'windows.storage.dll'},
    'api-ms-win-shell-shdirectory-l1-1-0': {None: 'shcore.dll'},
    'api-ms-win-shell-shellcom-l1-1-0': {None: 'kernelbase.dll'},
    'api-ms-win-shell-shellfolders-l1-1-0': {None: 'windows.storage.dll'},
    'api-ms-win-shlwapi-ie-l1-1-0': {None: 'shlwapi.dll'},
    'api-ms-win-shlwapi-winrt-storage-l1-1-1': {None: 'shlwapi.dll'},
    'api-ms-win-storage-exports-external-l1-1-0': {None: 'windows.storage.dll'},
    'api-ms-win-storage-exports-internal-l1-1-0': {None: 'windows.storage.dll'},
    'api-ms-win-winrt-search-folder-l1-1-0': {None: 'windows.storage.search.dll'},
    'ext-ms-mf-pal-l2-1-0': {None: ''},
    'ext-ms-net-eap-sim-l1-1-0': {None: 'eapsimextdesktop.dll'},
    'ext-ms-net-vpn-soh-l1-1-0': {None: 'vpnsohdesktop.dll'},
    'ext-ms-onecore-appchromeapi-l1-1-0': {None: ''},
    'ext-ms-onecore-appdefaults-l1-1-0': {None: 'windows.storage.dll'},
    'ext-ms-onecore-appmodel-emclient-l1-1-0': {None: ''},
    'ext-ms-onecore-appmodel-emsvcs-l1-1-0': {None: ''},
    'ext-ms-onecore-appmodel-pacmanclient-l1-1-0': {None: ''},
    'ext-ms-onecore-appmodel-plm-l1-1-1': {None: 'execmodelclient.dll'},
    'ext-ms-onecore-appmodel-staterepository-internal-l1-1-1': {None: 'windows.staterepositoryclient.dll'},
    'ext-ms-onecore-appmodel-tdlmigration-l1-1-0': {None: 'tdlmigration.dll'},
    'ext-ms-onecore-appmodel-veventdispatcher-l1-1-0': {None: 'veeventdispatcher.dll'},
    'ext-ms-onecore-comp-dwmmonitor-l1-1-0': {None: ''},
    'ext-ms-onecore-dcomp-l1-1-0': {None: 'dcomp.dll'},
    'ext-ms-onecore-defaultdiscovery-l1-1-0': {None: ''},
    'ext-ms-onecore-hcap-svf-l1-1-0': {None: 'svf.dll'},
    'ext-ms-onecore-hlink-l1-1-0': {None: 'hlink.dll'},
    'ext-ms-onecore-hnetcfg-l1-1-0': {None: 'hnetcfgclient.dll'},
    'ext-ms-onecore-mpc-input-l1-1-0': {None: 'shellcompositor.dll'},
    'ext-ms-onecore-orientation-l1-1-0': {None: ''},
    'ext-ms-onecore-phoneinfo-l1-1-0': {None: ''},
    'ext-ms-onecore-shellchromeapi-l1-1-2': {None: ''},
    'ext-ms-onecore-shellremindersapi-l1-1-0': {None: ''},
    'ext-ms-onecore-shlwapi-l1-1-0': {None: 'shlwapi.dll'},
    'ext-ms-onecore-spectrumsyncclient-l1-1-0': {None: 'spectrumsyncclient.dll'},
    'ext-ms-win-adsi-activeds-l1-1-0': {None: 'activeds.dll'},
    'ext-ms-win-advapi32-auth-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-encryptedfile-l1-1-1': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-eventlog-ansi-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-eventlog-l1-1-1': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-hwprof-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-idletask-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-lsa-l1-1-1': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-msi-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-npusername-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-ntmarta-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-psm-app-l1-1-0': {None: 'twinapi.appcore.dll'},
    'ext-ms-win-advapi32-registry-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-safer-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-advapi32-shutdown-l1-1-0': {None: 'advapi32.dll'},
    'ext-ms-win-appcompat-aepic-l1-1-0': {None: 'aepic.dll'},
    'ext-ms-win-appcompat-apphelp-l1-1-0': {None: 'apphelp.dll'},
    'ext-ms-win-appcompat-pcacli-l1-1-0': {None: 'pcacli.dll'},
    'ext-ms-win-appmodel-appcontainerpath-l1-1-0': {None: ''},
    'ext-ms-win-appmodel-datasharingservice-extensions-l1-1-0': {None: ''},
    'ext-ms-win-appmodel-daxcore-l1-1-1': {None: 'daxexec.dll'},
    'ext-ms-win-appmodel-deployment-l1-1-1': {None: ''},
    'ext-ms-win-appmodel-deploymentvolumes-l1-1-1': {None: ''},
    'ext-ms-win-appmodel-opc-l1-1-0': {None: 'opcservices.dll'},
    'ext-ms-win-appmodel-restrictedappcontainer-internal-l1-1-0': {None: 'kernel.appcore.dll'},
    'ext-ms-win-appmodel-state-ext-l1-2-0': {None: 'kernel.appcore.dll'},
    'ext-ms-win-appmodel-usercontext-l1-1-0': {None: ''},
    'ext-ms-win-appmodel-viewscalefactor-l1-1-0': {None: ''},
    'ext-ms-win-appxdeploymentclient-appxdeploy-l1-1-0': {None: 'appxdeploymentclient.dll'},
    'ext-ms-win-appxdeploymentclient-appxdeployonecore-l1-1-0': {None: 'appxdeploymentclient.dll'},
    'ext-ms-win-audiocore-coreaudiopolicymanager-l1-1-0': {None: 'coreaudiopolicymanagerext.dll'},
    'ext-ms-win-audiocore-pal-l1-2-0': {None: ''},
    'ext-ms-win-audiocore-policymanager-l1-1-0': {None: ''},
    'ext-ms-win-audiocore-spatial-l1-1-0': {None: ''},
    'ext-ms-win-authz-claimpolicies-l1-1-0': {None: 'authz.dll'},
    'ext-ms-win-authz-context-l1-1-0': {None: 'authz.dll'},
    'ext-ms-win-authz-remote-l1-1-0': {None: 'logoncli.dll'},
    'ext-ms-win-base-psapi-l1-1-0': {None: 'psapi.dll'},
    'ext-ms-win-base-rstrtmgr-l1-1-0': {None: 'rstrtmgr.dll'},
    'ext-ms-win-biometrics-winbio-core-l1-1-2': {None: 'winbio.dll'},
    'ext-ms-win-biometrics-winbio-l1-1-0': {None: 'winbio.dll'},
    'ext-ms-win-biometrics-winbio-l1-2-0': {None: 'winbioext.dll'},
    'ext-ms-win-biometrics-winbio-l1-3-0': {None: 'winbioext.dll'},
    'ext-ms-win-bluetooth-apis-l1-1-0': {None: 'bluetoothapis.dll'},
    'ext-ms-win-bluetooth-apis-private-l1-1-0': {None: 'bluetoothapis.dll'},
    'ext-ms-win-branding-winbrand-l1-1-1': {None: 'winbrand.dll'},
    'ext-ms-win-casting-device-l1-1-0': {None: ''},
    'ext-ms-win-casting-lockscreen-l1-1-0': {None: 'castingshellext.dll'},
    'ext-ms-win-casting-receiver-l1-1-1': {None: ''},
    'ext-ms-win-casting-shell-l1-1-0': {None: 'castingshellext.dll'},
    'ext-ms-win-ci-xbox-l1-1-0': {None: ''},
    'ext-ms-win-cloudap-tbal-l1-1-0': {None: ''},
    'ext-ms-win-clouddomainjoin-usermanagement-l1-1-0': {None: ''},
    'ext-ms-win-cluster-clusapi-l1-1-2': {None: 'clusapi.dll'},
    'ext-ms-win-cluster-resutils-l1-1-1': {None: 'resutils.dll'},
    'ext-ms-win-cmd-util-l1-1-0': {None: 'cmdext.dll'},
    'ext-ms-win-cng-rng-l1-1-1': {None: 'bcryptprimitives.dll'},
    'ext-ms-win-com-clbcatq-l1-1-0': {None: 'clbcatq.dll'},
    'ext-ms-win-com-coml2-l1-1-1': {None: 'coml2.dll'},
    'ext-ms-win-com-ole32-l1-1-5': {None: 'ole32.dll'},
    'ext-ms-win-com-psmregister-l1-1-0': {None: 'kernel.appcore.dll'},
    'ext-ms-win-com-psmregister-l1-2-2': {None: 'kernel.appcore.dll'},
    'ext-ms-win-com-suspendresiliency-l1-1-0': {None: ''},
    'ext-ms-win-composition-ghost-l1-1-0': {None: 'dwmghost.dll'},
    'ext-ms-win-composition-init-l1-1-0': {None: 'dwminit.dll'},
    'ext-ms-win-compositor-hosting-l1-1-1': {None: 'ism.dll'},
    'ext-ms-win-core-app-package-registration-l1-1-1': {None: ''},
    'ext-ms-win-core-app-package-volume-l1-1-0': {None: ''},
    'ext-ms-win-core-dhcp6client-l1-1-0': {None: ''},
    'ext-ms-win-core-iuri-l1-1-0': {None: 'urlmon.dll'},
    'ext-ms-win-core-licensemanager-l1-1-0': {None: ''},
    'ext-ms-win-core-psm-extendedresourcemode-l1-1-0': {None: ''},
    'ext-ms-win-core-psm-service-l1-1-3': {None: 'psmserviceexthost.dll'},
    'ext-ms-win-core-resourcemanager-l1-1-0': {None: 'rmclient.dll'},
    'ext-ms-win-core-resourcemanager-l1-2-0': {None: 'rmclient.dll'},
    'ext-ms-win-core-resourcepolicy-l1-1-2': {None: 'resourcepolicyclient.dll'},
    'ext-ms-win-core-resourcepolicyserver-l1-1-0': {None: 'resourcepolicyserver.dll'},
    'ext-ms-win-core-stateseparationext-l1-1-0': {None: ''},
    'ext-ms-win-core-storelicensing-l1-1-0': {None: 'licensemanagerapi.dll'},
    'ext-ms-win-core-storelicensing-l1-2-0': {None: 'licensemanagerapi.dll'},
    'ext-ms-win-core-winrt-remote-l1-1-0': {None: ''},
    'ext-ms-win-core-winsrv-l1-1-0': {None: 'winsrvext.dll'},
    'ext-ms-win-core-xbrm-l1-1-0': {None: ''},
    'ext-ms-win-coreui-l1-1-0': {None: ''},
    'ext-ms-win-coreui-navshutdown-l1-1-0': {None: 'navshutdown.dll'},
    'ext-ms-win-crypto-xbox-l1-1-0': {None: ''},
    'ext-ms-win-deployment-productenumerator-l1-1-0': {None: 'productenumerator.dll'},
    'ext-ms-win-desktopappx-l1-1-1': {None: 'daxexec.dll'},
    'ext-ms-win-devmgmt-dm-l1-1-1': {None: 'dmapisetextimpldesktop.dll'},
    'ext-ms-win-devmgmt-policy-l1-1-1': {None: 'policymanager.dll'},
    'ext-ms-win-direct2d-desktop-l1-1-0': {None: 'direct2ddesktop.dll'},
    'ext-ms-win-domainjoin-netjoin-l1-1-0': {None: 'netjoin.dll'},
    'ext-ms-win-dwmapi-ext-l1-1-0': {None: 'dwmapi.dll'},
    'ext-ms-win-dwmapidxgi-ext-l1-1-0': {None: 'dwmapi.dll'},
    'ext-ms-win-dx-d3d9-l1-1-0': {None: 'd3d9.dll'},
    'ext-ms-win-dx-ddraw-l1-1-0': {None: 'ddraw.dll'},
    'ext-ms-win-dx-dinput8-l1-1-0': {None: 'dinput8.dll'},
    'ext-ms-win-edputil-policy-l1-1-1': {None: 'edputil.dll'},
    'ext-ms-win-els-elscore-l1-1-0': {None: 'elscore.dll'},
    'ext-ms-win-eventing-pdh-l1-1-0': {None: 'pdh.dll'},
    'ext-ms-win-eventing-rundown-l1-1-0': {None: 'etwrundown.dll'},
    'ext-ms-win-eventing-tdh-ext-l1-1-0': {None: 'tdh.dll'},
    'ext-ms-win-familysafety-childaccount-l1-1-0': {None: 'familysafetyext.dll'},
    'ext-ms-win-feclient-encryptedfile-l1-1-1': {None: 'feclient.dll'},
    'ext-ms-win-firewallapi-webproxy-l1-1-1': {None: 'firewallapi.dll'},
    'ext-ms-win-font-fontgroups-l1-1-0': {None: 'fontgroupsoverride.dll'},
    'ext-ms-win-fs-clfs-l1-1-0': {None: 'clfs.sys'},
    'ext-ms-win-fs-cscapi-l1-1-0': {None: 'cscapi.dll'},
    'ext-ms-win-fs-vssapi-l1-1-0': {None: 'vssapi.dll'},
    'ext-ms-win-fsutilext-ifsutil-l1-1-0': {None: 'fsutilext.dll'},
    'ext-ms-win-fsutilext-ulib-l1-1-0': {None: 'fsutilext.dll'},
    'ext-ms-win-fveapi-query-l1-1-0': {None: 'fveapi.dll'},
    'ext-ms-win-gaming-gamechatoverlay-l1-1-0': {None: 'gamechatoverlayext.dll'},
    'ext-ms-win-gaming-xblgamesave-l1-1-0': {None: 'xblgamesaveext.dll'},
    'ext-ms-win-gaming-xinput-l1-1-0': {None: 'xinputuap.dll'},
    'ext-ms-win-gdi-clipping-l1-1-0': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-dc-create-l1-1-2': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-dc-l1-2-1': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-devcaps-l1-1-0': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-draw-l1-1-3': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-font-l1-1-3': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-gdiplus-l1-1-0': {None: 'gdiplus.dll'},
    'ext-ms-win-gdi-internal-desktop-l1-1-1': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-metafile-l1-1-2': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-path-l1-1-0': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-print-l1-1-0': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-private-l1-1-0': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-render-l1-1-0': {None: 'gdi32.dll'},
    'ext-ms-win-gdi-rgn-l1-1-0': {None: 'gdi32full.dll'},
    'ext-ms-win-gdi-wcs-l1-1-0': {None: 'gdi32full.dll'},
    'ext-ms-win-globalization-collation-l1-1-0': {None: 'globcollationhost.dll'},
    'ext-ms-win-globalization-input-l1-1-1': {None: 'globinputhost.dll'},
    'ext-ms-win-gpapi-grouppolicy-l1-1-0': {None: 'gpapi.dll'},
    'ext-ms-win-gpsvc-grouppolicy-l1-1-0': {None: 'gpsvc.dll'},
    'ext-ms-win-gui-dui70-l1-1-0': {None: 'dui70.dll'},
    'ext-ms-win-gui-ieui-l1-1-0': {None: 'ieui.dll'},
    'ext-ms-win-gui-uxinit-l1-1-0': {None: 'uxinit.dll'},
    'ext-ms-win-hyperv-compute-l1-1-0': {None: 'vmcompute.dll'},
    'ext-ms-win-imm-l1-1-1': {None: 'imm32.dll'},
    'ext-ms-win-kernel32-appcompat-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-datetime-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-elevation-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-errorhandling-l1-1-0': {None: 'kernel32.dll',
                                                 'kernel32.dll': 'faultrep.dll'},
    'ext-ms-win-kernel32-file-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-localization-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-package-current-l1-1-0': {None: 'kernel.appcore.dll'},
    'ext-ms-win-kernel32-package-l1-1-2': {None: 'kernel.appcore.dll'},
    'ext-ms-win-kernel32-process-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-quirks-l1-1-1': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-registry-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-sidebyside-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-transacted-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kernel32-windowserrorreporting-l1-1-1': {None: 'kernel32.dll'},
    'ext-ms-win-kernelbase-processthread-l1-1-0': {None: 'kernel32.dll'},
    'ext-ms-win-kioskmode-config-l1-1-0': {None: 'lockdownutil.dll'},
    'ext-ms-win-mapi-mapi32-l1-1-0': {None: 'mapistub.dll'},
    'ext-ms-win-media-avi-l1-1-0': {None: 'avifil32.dll'},
    'ext-ms-win-mf-pal-l1-1-0': {None: ''},
    'ext-ms-win-mf-vfw-l1-1-0': {None: 'mfvfw.dll'},
    'ext-ms-win-mininput-cursorhost-l1-1-0': {None: 'inputhost.dll'},
    'ext-ms-win-mininput-extensions-l1-1-0': {None: ''},
    'ext-ms-win-mininput-inputhost-l1-1-0': {None: 'inputhost.dll'},
    'ext-ms-win-mininput-systeminputhost-l1-1-0': {None: 'ism.dll'},
    'ext-ms-win-mm-io-l1-1-0': {None: ''},
    'ext-ms-win-mm-msacm-l1-1-0': {None: 'msacm32.dll'},
    'ext-ms-win-mm-pehelper-l1-1-0': {None: 'mf.dll'},
    'ext-ms-win-mm-wmvcore-l1-1-0': {None: 'wmvcore.dll'},
    'ext-ms-win-mobilecore-boot-l1-1-0': {None: ''},
    'ext-ms-win-mobilecore-deviceinfo-l1-1-0': {None: ''},
    'ext-ms-win-mobilecore-ie-textinput-l1-1-0': {None: ''},
    'ext-ms-win-moderncore-win32k-base-ntgdi-l1-1-0': {None: 'win32kfull.sys'},
    'ext-ms-win-moderncore-win32k-base-ntuser-l1-1-0': {None: 'win32kfull.sys'},
    'ext-ms-win-moderncore-win32k-base-sysentry-l1-1-0': {None: 'win32k.sys'},
    'ext-ms-win-mpr-multipleproviderrouter-l1-1-0': {None: 'mprext.dll'},
    'ext-ms-win-mrmcorer-environment-l1-1-0': {None: ''},
    'ext-ms-win-mrmcorer-resmanager-l1-1-0': {None: 'mrmcorer.dll'},
    'ext-ms-win-msa-device-l1-1-0': {None: ''},
    'ext-ms-win-msa-ui-l1-1-0': {None: 'msauserext.dll'},
    'ext-ms-win-msa-user-l1-1-1': {None: 'msauserext.dll'},
    'ext-ms-win-msi-misc-l1-1-0': {None: 'msi.dll'},
    'ext-ms-win-msiltcfg-msi-l1-1-0': {None: 'msiltcfg.dll'},
    'ext-ms-win-msimg-draw-l1-1-0': {None: 'msimg32.dll'},
    'ext-ms-win-net-cmvpn-l1-1-0': {None: 'cmintegrator.dll'},
    'ext-ms-win-net-httpproxyext-l1-1-0': {None: 'httpprxc.dll'},
    'ext-ms-win-net-isoext-l1-1-0': {None: 'firewallapi.dll'},
    'ext-ms-win-net-netbios-l1-1-0': {None: 'netbios.dll'},
    'ext-ms-win-net-netshell-l1-1-0': {None: 'netshell.dll'},
    'ext-ms-win-net-nfdapi-l1-1-0': {None: 'ndfapi.dll'},
    'ext-ms-win-net-vpn-l1-1-0': {None: ''},
    'ext-ms-win-netprovision-netprovfw-l1-1-0': {None: 'netprovfw.dll'},
    'ext-ms-win-networking-iphlpsvc-l1-1-0': {None: ''},
    'ext-ms-win-networking-mpssvc-l1-1-0': {None: ''},
    'ext-ms-win-networking-ncsiuserprobe-l1-1-0': {None: ''},
    'ext-ms-win-networking-nlaapi-l1-1-0': {None: 'nlaapi.dll'},
    'ext-ms-win-networking-radiomonitor-l1-1-0': {None: 'windows.devices.radios.dll'},
    'ext-ms-win-networking-wcmapi-l1-1-0': {None: 'wcmapi.dll'},
    'ext-ms-win-networking-winipsec-l1-1-0': {None: 'winipsec.dll'},
    'ext-ms-win-networking-wlanapi-l1-1-0': {None: 'wlanapi.dll'},
    'ext-ms-win-networking-wlanstorage-l1-1-0': {None: ''},
    'ext-ms-win-networking-xblconnectivity-l1-1-0': {None: ''},
    'ext-ms-win-newdev-config-l1-1-2': {None: 'newdev.dll'},
    'ext-ms-win-nfc-semgr-l1-1-0': {None: 'semgrsvc.dll'},
    'ext-ms-win-ntdsa-activedirectoryserver-l1-1-0': {None: 'ntdsa.dll'},
    'ext-ms-win-ntdsapi-activedirectoryclient-l1-1-1': {None: 'ntdsapi.dll'},
    'ext-ms-win-ntos-clipsp-l1-1-0': {None: 'clipsp.sys'},
    'ext-ms-win-ntos-ium-l1-1-0': {None: ''},
    'ext-ms-win-ntos-kcminitcfg-l1-1-0': {None: 'cmimcext.sys'},
    'ext-ms-win-ntos-ksecurity-l1-1-1': {None: ''},
    'ext-ms-win-ntos-ksigningpolicy-l1-1-0': {None: ''},
    'ext-ms-win-ntos-ksr-l1-1-2': {None: ''},
    'ext-ms-win-ntos-tm-l1-1-0': {None: 'tm.sys'},
    'ext-ms-win-ntos-ucode-l1-1-0': {None: 'ntosext.sys'},
    'ext-ms-win-ntos-werkernel-l1-1-1': {None: 'werkernel.sys'},
    'ext-ms-win-ntuser-caret-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-chartranslation-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-dc-access-ext-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-dialogbox-l1-1-3': {None: 'user32.dll'},
    'ext-ms-win-ntuser-draw-l1-1-2': {None: 'user32.dll'},
    'ext-ms-win-ntuser-gui-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-gui-l1-2-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-gui-l1-3-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-keyboard-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-keyboard-l1-2-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-keyboard-l1-3-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-menu-l1-1-3': {None: 'user32.dll'},
    'ext-ms-win-ntuser-message-l1-1-3': {None: 'user32.dll'},
    'ext-ms-win-ntuser-misc-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-misc-l1-2-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-misc-l1-3-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-misc-l1-5-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-mit-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-mouse-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-powermanagement-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-private-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-private-l1-2-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-private-l1-3-2': {None: 'user32.dll'},
    'ext-ms-win-ntuser-rawinput-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-rectangle-ext-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-rim-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-rotationmanager-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-server-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-string-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-synch-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-sysparams-ext-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-touch-hittest-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-uicontext-ext-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-ntuser-window-l1-1-4': {None: 'user32.dll'},
    'ext-ms-win-ntuser-windowclass-l1-1-2': {None: 'user32.dll'},
    'ext-ms-win-ntuser-windowstation-ansi-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-ntuser-windowstation-l1-1-2': {None: 'user32.dll'},
    'ext-ms-win-odbc-odbc32-l1-1-0': {None: 'odbc32.dll'},
    'ext-ms-win-ole32-bindctx-l1-1-0': {None: 'ole32.dll'},
    'ext-ms-win-ole32-ie-ext-l1-1-0': {None: 'ole32.dll'},
    'ext-ms-win-ole32-oleautomation-l1-1-0': {None: 'ole32.dll'},
    'ext-ms-win-oleacc-l1-1-2': {None: 'oleacc.dll'},
    'ext-ms-win-onecore-shutdown-l1-1-0': {None: ''},
    'ext-ms-win-parentalcontrols-setup-l1-1-0': {None: 'wpcapi.dll'},
    'ext-ms-win-pinenrollment-enrollment-l1-1-1': {None: 'pinenrollmenthelper.dll'},
    'ext-ms-win-printer-prntvpt-l1-1-1': {None: 'prntvpt.dll'},
    'ext-ms-win-printer-winspool-l1-1-4': {None: 'winspool.drv'},
    'ext-ms-win-profile-extender-l1-1-0': {None: 'userenv.dll'},
    'ext-ms-win-profile-load-l1-1-0': {None: ''},
    'ext-ms-win-profile-profsvc-l1-1-0': {None: 'profsvcext.dll'},
    'ext-ms-win-profile-userenv-l1-1-1': {None: 'profext.dll'},
    'ext-ms-win-provisioning-platform-l1-1-0': {None: 'provplatformdesktop.dll'},
    'ext-ms-win-ras-rasapi32-l1-1-2': {None: 'rasapi32.dll'},
    'ext-ms-win-ras-rasdlg-l1-1-0': {None: 'rasdlg.dll'},
    'ext-ms-win-ras-rasman-l1-1-0': {None: 'rasman.dll'},
    'ext-ms-win-ras-tapi32-l1-1-1': {None: 'tapi32.dll'},
    'ext-ms-win-raschapext-eap-l1-1-0': {None: 'raschapext.dll'},
    'ext-ms-win-rastlsext-eap-l1-1-0': {None: 'rastlsext.dll'},
    'ext-ms-win-rdr-davhlpr-l1-1-0': {None: 'davhlpr.dll'},
    'ext-ms-win-reinfo-query-l1-1-0': {None: 'reinfo.dll'},
    'ext-ms-win-remotewipe-platform-l1-1-0': {None: ''},
    'ext-ms-win-resourcemanager-gamemode-l1-1-0': {None: 'rmclient.dll'},
    'ext-ms-win-resourcemanager-gamemode-l1-2-0': {None: 'rmclient.dll'},
    'ext-ms-win-resources-deployment-l1-1-0': {None: 'mrmdeploy.dll'},
    'ext-ms-win-ro-typeresolution-l1-1-0': {None: 'wintypes.dll'},
    'ext-ms-win-rometadata-dispenser-l1-1-0': {None: 'rometadata.dll'},
    'ext-ms-win-rpc-firewallportuse-l1-1-0': {None: 'rpcrtremote.dll'},
    'ext-ms-win-rpc-ssl-l1-1-0': {None: 'rpcrtremote.dll'},
    'ext-ms-win-rtcore-gdi-devcaps-l1-1-1': {None: 'gdi32.dll'},
    'ext-ms-win-rtcore-gdi-object-l1-1-0': {None: 'gdi32.dll'},
    'ext-ms-win-rtcore-gdi-rgn-l1-1-1': {None: 'gdi32.dll'},
    'ext-ms-win-rtcore-minuser-host-l1-1-0': {None: ''},
    'ext-ms-win-rtcore-minuser-input-l1-1-2': {None: ''},
    'ext-ms-win-rtcore-minuser-private-ext-l1-1-0': {None: ''},
    'ext-ms-win-rtcore-ntuser-console-l1-1-0': {None: ''},
    'ext-ms-win-rtcore-ntuser-controllernavigation-l1-1-0': {None: ''},
    'ext-ms-win-rtcore-ntuser-cursor-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-dc-access-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-dpi-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-dpi-l1-2-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-iam-l1-1-1': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-inputintercept-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-integration-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-mininit-l1-1-0': {None: ''},
    'ext-ms-win-rtcore-ntuser-rawinput-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-synch-ext-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-syscolors-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-sysparams-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-window-ext-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-winevent-ext-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-rtcore-ntuser-wmpointer-l1-1-0': {None: 'user32.dll'},
    'ext-ms-win-samsrv-accountstore-l1-1-0': {None: 'samsrv.dll'},
    'ext-ms-win-scesrv-server-l1-1-0': {None: 'scesrv.dll'},
    'ext-ms-win-search-folder-l1-1-0': {None: 'searchfolder.dll'},
    'ext-ms-win-secur32-translatename-l1-1-1': {None: 'secur32.dll'},
    'ext-ms-win-security-appinfoext-l1-1-0': {None: 'appinfoext.dll'},
    'ext-ms-win-security-authbrokerui-l1-1-0': {None: 'authbrokerui.dll'},
    'ext-ms-win-security-capauthz-ext-l1-1-0': {None: ''},
    'ext-ms-win-security-capauthz-l1-1-1': {None: 'capauthz.dll'},
    'ext-ms-win-security-certpoleng-l1-1-0': {None: 'certpoleng.dll'},
    'ext-ms-win-security-chambers-l1-1-1': {None: ''},
    'ext-ms-win-security-credui-internal-l1-1-0': {None: 'wincredui.dll'},
    'ext-ms-win-security-credui-l1-1-1': {None: 'credui.dll'},
    'ext-ms-win-security-cryptui-l1-1-1': {None: 'cryptui.dll'},
    'ext-ms-win-security-developerunlock-l1-1-0': {None: ''},
    'ext-ms-win-security-deviceid-l1-1-0': {None: ''},
    'ext-ms-win-security-efs-l1-1-1': {None: 'efsext.dll'},
    'ext-ms-win-security-efswrt-l1-1-1': {None: 'efswrt.dll'},
    'ext-ms-win-security-kerberos-l1-1-0': {None: 'kerberos.dll'},
    'ext-ms-win-security-shutdownext-l1-1-0': {None: 'shutdownext.dll'},
    'ext-ms-win-security-slc-l1-1-0': {None: 'slc.dll'},
    'ext-ms-win-security-srp-l1-1-1': {None: 'srpapi.dll'},
    'ext-ms-win-security-tokenbrokerui-l1-1-0': {None: 'tokenbrokerui.dll'},
    'ext-ms-win-security-vaultcli-l1-1-1': {None: 'vaultcli.dll'},
    'ext-ms-win-security-winscard-l1-1-0': {None: 'winscard.dll'},
    'ext-ms-win-sensors-core-private-l1-1-2': {None: 'sensorsnativeapi.dll'},
    'ext-ms-win-sensors-utilities-private-l1-1-1': {None: 'sensorsutilsv2.dll'},
    'ext-ms-win-session-userinit-l1-1-0': {None: 'userinitext.dll'},
    'ext-ms-win-session-usermgr-l1-1-0': {None: 'usermgrcli.dll'},
    'ext-ms-win-session-usermgr-l1-2-0': {None: 'usermgrcli.dll'},
    'ext-ms-win-session-usertoken-l1-1-0': {None: 'wtsapi32.dll'},
    'ext-ms-win-session-wininit-l1-1-0': {None: 'wininitext.dll'},
    'ext-ms-win-session-winlogon-l1-1-0': {None: 'winlogonext.dll'},
    'ext-ms-win-session-winsta-l1-1-2': {None: 'winsta.dll'},
    'ext-ms-win-session-wtsapi32-l1-1-0': {None: 'wtsapi32.dll'},
    'ext-ms-win-setupapi-classinstallers-l1-1-2': {None: 'setupapi.dll'},
    'ext-ms-win-setupapi-inf-l1-1-1': {None: 'setupapi.dll'},
    'ext-ms-win-setupapi-logging-l1-1-0': {None: 'setupapi.dll'},
    'ext-ms-win-shell-aclui-l1-1-0': {None: 'aclui.dll'},
    'ext-ms-win-shell-browsersettingsync-l1-1-0': {None: ''},
    'ext-ms-win-shell-comctl32-da-l1-1-0': {None: 'comctl32.dll'},
    'ext-ms-win-shell-comctl32-init-l1-1-1': {None: 'comctl32.dll'},
    'ext-ms-win-shell-comctl32-l1-1-0': {None: 'comctl32.dll'},
    'ext-ms-win-shell-comctl32-window-l1-1-0': {None: 'comctl32.dll'},
    'ext-ms-win-shell-comdlg32-l1-1-1': {None: 'comdlg32.dll'},
    'ext-ms-win-shell-directory-l1-1-0': {None: 'windows.storage.dll'},
    'ext-ms-win-shell-efsadu-l1-1-0': {None: 'efsadu.dll'},
    'ext-ms-win-shell-embeddedmode-l1-1-0': {None: 'embeddedmodesvcapi.dll'},
    'ext-ms-win-shell-exports-internal-l1-1-0': {None: 'shell32.dll'},
    'ext-ms-win-shell-knownfolderext-l1-1-0': {None: ''},
    'ext-ms-win-shell-ntshrui-l1-1-0': {None: 'ntshrui.dll'},
    'ext-ms-win-shell-propsys-l1-1-0': {None: 'propsys.dll'},
    'ext-ms-win-shell-settingsync-l1-1-3': {None: 'settingsyncpolicy.dll'},
    'ext-ms-win-shell-shdocvw-l1-1-0': {None: 'shdocvw.dll'},
    'ext-ms-win-shell-shell32-l1-2-2': {None: 'shell32.dll'},
    'ext-ms-win-shell-shlwapi-l1-1-2': {None: 'shlwapi.dll'},
    'ext-ms-win-shell-shlwapi-l1-2-0': {None: 'shlwapi.dll'},
    'ext-ms-win-shell32-shellcom-l1-1-0': {None: 'windows.storage.dll'},
    'ext-ms-win-shell32-shellfolders-l1-1-1': {None: 'windows.storage.dll'},
    'ext-ms-win-shell32-shellfolders-l1-2-0': {None: 'windows.storage.dll'},
    'ext-ms-win-smbshare-browser-l1-1-0': {None: 'browser.dll'},
    'ext-ms-win-smbshare-browserclient-l1-1-0': {None: 'browcli.dll'},
    'ext-ms-win-smbshare-sscore-l1-1-0': {None: 'sscoreext.dll'},
    'ext-ms-win-spinf-inf-l1-1-0': {None: 'spinf.dll'},
    'ext-ms-win-storage-hbaapi-l1-1-0': {None: 'hbaapi.dll'},
    'ext-ms-win-storage-iscsidsc-l1-1-0': {None: 'iscsidsc.dll'},
    'ext-ms-win-storage-sense-l1-1-0': {None: 'storageusage.dll'},
    'ext-ms-win-storage-sense-l1-2-0': {None: 'storageusage.dll'},
    'ext-ms-win-sxs-oleautomation-l1-1-0': {None: 'sxs.dll'},
    'ext-ms-win-test-sys1-l1-1-0': {None: ''},
    'ext-ms-win-test-sys2-l1-1-0': {None: ''},
    'ext-ms-win-tsf-inputsetting-l1-1-0': {None: 'input.dll'},
    'ext-ms-win-tsf-msctf-l1-1-2': {None: 'msctf.dll'},
    'ext-ms-win-ttlsext-eap-l1-1-0': {None: 'ttlsext.dll'},
    'ext-ms-win-ui-viewmanagement-l1-1-0': {None: ''},
    'ext-ms-win-uiacore-l1-1-3': {None: 'uiautomationcore.dll'},
    'ext-ms-win-umpoext-umpo-l1-1-0': {None: 'umpoext.dll'},
    'ext-ms-win-usp10-l1-1-0': {None: 'gdi32full.dll'},
    'ext-ms-win-uwf-servicing-apis-l1-1-0': {None: 'uwfservicingapi.dll'},
    'ext-ms-win-uxtheme-themes-l1-1-1': {None: 'uxtheme.dll'},
    'ext-ms-win-wer-reporting-l1-1-2': {None: 'wer.dll'},
    'ext-ms-win-wer-ui-l1-1-0': {None: 'werui.dll'},
    'ext-ms-win-wer-wct-l1-1-0': {None: 'wer.dll'},
    'ext-ms-win-wer-xbox-l1-1-0': {None: ''},
    'ext-ms-win-wevtapi-eventlog-l1-1-3': {None: 'wevtapi.dll'},
    'ext-ms-win-winlogon-mincreds-l1-1-0': {None: ''},
    'ext-ms-win-winrt-networking-connectivity-l1-1-0': {None: ''},
    'ext-ms-win-winrt-storage-fileexplorer-l1-1-0': {None: ''},
    'ext-ms-win-winrt-storage-l1-1-0': {None: 'windows.storage.dll'},
    'ext-ms-win-winrt-storage-removable-l1-1-0': {None: ''},
    'ext-ms-win-winrt-storage-win32broker-l1-1-0': {None: 'windows.storage.winrt.dll'},
    'ext-ms-win-wlan-grouppolicy-l1-1-0': {None: 'wlgpclnt.dll'},
    'ext-ms-win-wlan-onexui-l1-1-0': {None: 'onexui.dll'},
    'ext-ms-win-wlan-scard-l1-1-0': {None: 'winscard.dll'},
    'ext-ms-win-wnv-l1-1-0': {None: 'wnv.sys'},
    'ext-ms-win-wpn-phoneext-l1-1-0': {None: ''},
    'ext-ms-win-wrp-sfc-l1-1-0': {None: 'sfc.dll'},
    'ext-ms-win-wsclient-devlicense-l1-1-1': {None: 'wsclient.dll'},
    'ext-ms-win-wwaext-misc-l1-1-0': {None: 'wwaext.dll'},
    'ext-ms-win-wwaext-module-l1-1-0': {None: 'wwaext.dll'},
    'ext-ms-win-wwan-wwapi-l1-1-2': {None: 'wwapi.dll'},
    'ext-ms-win-xaml-controls-l1-1-0': {None: 'windows.ui.xaml.phone.dll'},
    'ext-ms-win-xaml-pal-l1-1-0': {None: ''},
    'ext-ms-win-xblauth-console-l1-1-0': {None: ''},
    'ext-ms-win-xboxlive-xboxnetapisvc-l1-1-0': {None: ''}
}



#-*- coding:utf-8 -*-

from builtins import range

import logging
from pyparsing import *
from miasm.expression.expression import *
from miasm.core.cpu import *
from collections import defaultdict
from miasm.core.bin_stream import bin_stream
import miasm.arch.msp430.regs as regs_module
from miasm.arch.msp430.regs import *
from miasm.core.asm_ast import AstInt, AstId, AstMem, AstOp
from miasm.ir.ir import color_expr_html

log = logging.getLogger("msp430dis")
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter("[%(levelname)-8s]: %(message)s"))
log.addHandler(console_handler)
log.setLevel(logging.DEBUG)

conditional_branch = ['jnz', 'jz', 'jnc', 'jc',
                      'jn', 'jge', 'jl']
unconditional_branch = ['jmp']

def cb_deref_nooff(tokens):
    assert len(tokens) == 1
    result = AstMem(tokens[0], 16)
    return result


def cb_deref_pinc(tokens):
    assert len(tokens) == 1

    result = AstOp('autoinc', *tokens)
    return result


def cb_deref_off(tokens):
    assert len(tokens) == 2
    result = AstMem(tokens[1] + tokens[0], 16)
    return result


def cb_expr(tokens):
    assert(len(tokens) == 1)
    result = tokens[0]
    return result


ARO = Suppress("@")
LPARENT = Suppress("(")
RPARENT = Suppress(")")

PINC = Suppress("+")

deref_nooff = (ARO + base_expr).setParseAction(cb_deref_nooff)
deref_pinc = (ARO + base_expr + PINC).setParseAction(cb_deref_pinc)
deref_off = (base_expr + LPARENT + gpregs.parser + RPARENT).setParseAction(cb_deref_off)
sreg_p = (deref_pinc | deref_nooff | deref_off | base_expr).setParseAction(cb_expr)



class msp430_arg(m_arg):
    def asm_ast_to_expr(self, value, loc_db):
        if isinstance(value, AstId):
            name = value.name
            if is_expr(name):
                return name
            assert isinstance(name, str)
            if name in gpregs.str:
                index = gpregs.str.index(name)
                reg = gpregs.expr[index]
                return reg
            loc_key = loc_db.get_or_create_name_location(value.name)
            return ExprLoc(loc_key, 16)
        if isinstance(value, AstOp):
            args = [self.asm_ast_to_expr(tmp, loc_db) for tmp in value.args]
            if None in args:
                return None
            return ExprOp(value.op, *args)
        if isinstance(value, AstInt):
            return ExprInt(value.value, 16)
        if isinstance(value, AstMem):
            ptr = self.asm_ast_to_expr(value.ptr, loc_db)
            if ptr is None:
                return None
            return ExprMem(ptr, value.size)
        return None


class additional_info(object):

    def __init__(self):
        self.except_on_instr = False


class instruction_msp430(instruction):
    __slots__ = []

    def dstflow(self):
        if self.name.startswith('j'):
            return True
        return self.name in ['call']

    @staticmethod
    def arg2str(expr, index=None, loc_db=None):
        if isinstance(expr, ExprId):
            o = str(expr)
        elif isinstance(expr, ExprInt):
            o = str(expr)
        elif expr.is_loc():
            if loc_db is not None:
                return loc_db.pretty_str(expr.loc_key)
            else:
                return str(expr)
        elif isinstance(expr, ExprOp) and expr.op == "autoinc":
            o = "@%s+" % str(expr.args[0])
        elif isinstance(expr, ExprMem):
            if isinstance(expr.ptr, ExprId):
                if index == 0:
                    o = "@%s" % expr.ptr
                else:
                    o = "0x0(%s)" % expr.ptr
            elif isinstance(expr.ptr, ExprInt):
                o = "@%s" % expr.ptr
            elif isinstance(expr.ptr, ExprOp):
                o = "%s(%s)" % (expr.ptr.args[1], expr.ptr.args[0])
        else:
            raise NotImplementedError('unknown instance expr = %s' % type(expr))
        return o

    @staticmethod
    def arg2html(expr, index=None, loc_db=None):
        if isinstance(expr, ExprId) or isinstance(expr, ExprInt) or expr.is_loc():
            return color_expr_html(expr, loc_db)
        elif isinstance(expr, ExprOp) and expr.op == "autoinc":
            o = "@%s+" % color_expr_html(expr.args[0], loc_db)
        elif isinstance(expr, ExprMem):
            if isinstance(expr.ptr, ExprId):
                if index == 0:
                    o = "@%s" % color_expr_html(expr.ptr, loc_db)
                else:
                    o = "0x0(%s)" % color_expr_html(expr.ptr, loc_db)
            elif isinstance(expr.ptr, ExprInt):
                o = "@%s" % color_expr_html(expr.ptr, loc_db)
            elif isinstance(expr.ptr, ExprOp):
                o = "%s(%s)" % (
                    color_expr_html(expr.ptr.args[1], loc_db),
                    color_expr_html(expr.ptr.args[0], loc_db)
                )
        else:
            raise NotImplementedError('unknown instance expr = %s' % type(expr))
        return o


    def dstflow2label(self, loc_db):
        expr = self.args[0]
        if not isinstance(expr, ExprInt):
            return
        if self.name == "call":
            addr = int(expr)
        else:
            addr = (int(expr) + int(self.offset))  & int(expr.mask)

        loc_key = loc_db.get_or_create_offset_location(addr)
        self.args[0] = ExprLoc(loc_key, expr.size)

    def breakflow(self):
        if self.name in conditional_branch + unconditional_branch:
            return True
        if self.name.startswith('ret'):
            return True
        if self.name.startswith('int'):
            return True
        if self.name.startswith('mov') and self.args[1] == PC:
            return True
        return self.name in ['call']

    def splitflow(self):
        if self.name in conditional_branch:
            return True
        if self.name in unconditional_branch:
            return False
        return self.name in ['call']

    def setdstflow(self, a):
        return

    def is_subcall(self):
        return self.name in ['call']

    def getdstflow(self, loc_db):
        return [self.args[0]]

    def get_symbol_size(self, symbol, loc_db):
        return 16

    def fixDstOffset(self):
        e = self.args[0]
        if self.offset is None:
            raise ValueError('symbol not resolved %s' % l)
        if not isinstance(e, ExprInt):
            # raise ValueError('dst must be int or label')
            log.warning('dynamic dst %r', e)
            return

        # Call argument is an absolute offset
        # Other offsets are relative to instruction offset
        if self.name != "call":
            self.args[0] =  ExprInt(int(e) - self.offset, 16)

    def get_info(self, c):
        pass

    def __str__(self):
        o = super(instruction_msp430, self).__str__()
        return o

    def get_args_expr(self):
        args = []
        for a in self.args:
            args.append(a)
        return args


mode_msp430 = None


class mn_msp430(cls_mn):
    name = "msp430"
    regs = regs_module
    all_mn = []
    bintree = {}
    num = 0
    delayslot = 0
    pc = {None: PC}
    sp = {None: SP}
    all_mn_mode = defaultdict(list)
    all_mn_name = defaultdict(list)
    all_mn_inst = defaultdict(list)
    instruction = instruction_msp430
    max_instruction_len = 8

    @classmethod
    def getpc(cls, attrib):
        return PC

    @classmethod
    def getsp(cls, attrib):
        return SP

    @classmethod
    def check_mnemo(cls, fields):
        l = sum([x.l for x in fields])
        assert l % 16 == 00, "len %r" % l

    @classmethod
    def getbits(cls, bs, attrib, start, n):
        if not n:
            return 0
        o = 0
        if n > bs.getlen() * 8:
            raise ValueError('not enough bits %r %r' % (n, len(bs.bin) * 8))
        while n:
            i = start // 8
            c = cls.getbytes(bs, i)
            if not c:
                raise IOError
            c = ord(c)
            r = 8 - start % 8
            c &= (1 << r) - 1
            l = min(r, n)
            c >>= (r - l)
            o <<= l
            o |= c
            n -= l
            start += l
        return o

    @classmethod
    def getbytes(cls, bs, offset, l=1):
        out = b""
        for _ in range(l):
            n_offset = (offset & ~1) + 1 - offset % 2
            out += bs.getbytes(n_offset, 1)
            offset += 1
        return out

    def decoded2bytes(self, result):
        tmp = super(mn_msp430, self).decoded2bytes(result)
        out = []
        for x in tmp:
            o = b""
            while x:
                o += x[:2][::-1]
                x = x[2:]
            out.append(o)
        return out

    @classmethod
    def gen_modes(cls, subcls, name, bases, dct, fields):
        dct['mode'] = None
        return [(subcls, name, bases, dct, fields)]

    def additional_info(self):
        info = additional_info()
        return info

    @classmethod
    def getmn(cls, name):
        return name.upper()

    def reset_class(self):
        super(mn_msp430, self).reset_class()

    def getnextflow(self, loc_db):
        raise NotImplementedError('not fully functional')


def addop(name, fields, args=None, alias=False):
    dct = {"fields": fields}
    dct["alias"] = alias
    if args is not None:
        dct['args'] = args
    type(name, (mn_msp430,), dct)


class bw_mn(bs_mod_name):
    prio = 5
    mn_mod = ['.w', '.b']


class msp430_sreg_arg(reg_noarg, msp430_arg):
    prio = default_prio + 1
    reg_info = gpregs
    parser = sreg_p

    def decode(self, v):
        size = 16
        if hasattr(self.parent, 'size'):
            size = [16, 8][self.parent.size.value]
        v = v & self.lmask
        e = self.reg_info.expr[v]
        if self.parent.a_s.value == 0b00:
            if e == R3:
                self.expr = ExprInt(0, size)
            else:
                self.expr = e
        elif self.parent.a_s.value == 0b01:
            if e == SR:
                self.expr = ExprMem(ExprInt(self.parent.off_s.value, 16), size)
            elif e == R3:
                self.expr = ExprInt(1, size)
            else:
                self.expr = ExprMem(
                    e + ExprInt(self.parent.off_s.value, 16), size)
        elif self.parent.a_s.value == 0b10:
            if e == SR:
                self.expr = ExprInt(4, size)
            elif e == R3:
                self.expr = ExprInt(2, size)
            else:
                self.expr = ExprMem(e, size)
        elif self.parent.a_s.value == 0b11:
            if e == SR:
                self.expr = ExprInt(8, size)
            elif e == R3:
                if self.parent.size.value == 0:
                    self.expr = ExprInt(0xffff, size)
                else:
                    self.expr = ExprInt(0xff, size)
            elif e == PC:
                self.expr = ExprInt(self.parent.off_s.value, size)
            else:
                self.expr = ExprOp('autoinc', e)
        else:
            raise NotImplementedError(
                "unknown value self.parent.a_s.value = " +
                "%d" % self.parent.a_s.value)
        return True

    def encode(self):
        e = self.expr
        if e in self.reg_info.expr:
            self.parent.a_s.value = 0
            self.value = self.reg_info.expr.index(e)
        elif isinstance(e, ExprInt):
            v = int(e)
            if v == 0xffff and self.parent.size.value == 0:
                self.parent.a_s.value = 0b11
                self.value = 3
            elif v == 0xff and self.parent.size.value == 1:
                self.parent.a_s.value = 0b11
                self.value = 3
            elif v == 2:
                self.parent.a_s.value = 0b10
                self.value = 3
            elif v == 1:
                self.parent.a_s.value = 0b01
                self.value = 3
            elif v == 8:
                self.parent.a_s.value = 0b11
                self.value = 2
            elif v == 4:
                self.parent.a_s.value = 0b10
                self.value = 2
            elif v == 0:
                self.parent.a_s.value = 0b00
                self.value = 3
            else:
                self.parent.a_s.value = 0b11
                self.value = 0
                self.parent.off_s.value = v
        elif isinstance(e, ExprMem):
            if isinstance(e.ptr, ExprId):
                self.parent.a_s.value = 0b10
                self.value = self.reg_info.expr.index(e.ptr)
            elif isinstance(e.ptr, ExprInt):
                self.parent.a_s.value = 0b01
                self.value = self.reg_info.expr.index(SR)
                self.parent.off_s.value = int(e.ptr)
            elif isinstance(e.ptr, ExprOp):
                self.parent.a_s.value = 0b01
                self.value = self.reg_info.expr.index(e.ptr.args[0])
                self.parent.off_s.value = int(e.ptr.args[1])
            else:
                raise NotImplementedError(
                    'unknown instance e.ptr = %s' % type(e.ptr))
        elif isinstance(e, ExprOp) and e.op == "autoinc":
            self.parent.a_s.value = 0b11
            self.value = self.reg_info.expr.index(e.args[0])
        else:
            raise NotImplementedError('unknown instance e = %s' % type(e))
        return True


class msp430_dreg_arg(msp430_sreg_arg):
    prio = default_prio + 1
    reg_info = gpregs
    parser = sreg_p

    def decode(self, v):
        if hasattr(self.parent, 'size'):
            size = [16, 8][self.parent.size.value]
        else:
            size = 16

        v = v & self.lmask
        e = self.reg_info.expr[v]
        if self.parent.a_d.value == 0:
            self.expr = e
        elif self.parent.a_d.value == 1:
            if e == SR:
                x = ExprInt(self.parent.off_d.value, 16)
            else:
                x = e + ExprInt(self.parent.off_d.value, 16)
            self.expr = ExprMem(x, size)
        else:
            raise NotImplementedError(
                "unknown value self.parent.a_d.value = " +
                "%d" % self.parent.a_d.value)
        return True

    def encode(self):
        e = self.expr
        if e in self.reg_info.expr:
            self.parent.a_d.value = 0
            self.value = self.reg_info.expr.index(e)
        elif isinstance(e, ExprMem):
            if isinstance(e.ptr, ExprId):
                r, i = e.ptr, ExprInt(0, 16)
            elif isinstance(e.ptr, ExprOp):
                r, i = e.ptr.args[0], e.ptr.args[1]
            elif isinstance(e.ptr, ExprInt):
                r, i = SR, e.ptr
            else:
                raise NotImplementedError(
                    'unknown instance e.arg = %s' % type(e.ptr))
            self.parent.a_d.value = 1
            self.value = self.reg_info.expr.index(r)
            self.parent.off_d.value = int(i)
        else:
            raise NotImplementedError('unknown instance e = %s' % type(e))
        return True

class bs_cond_off_s(bs_cond):

    @classmethod
    def flen(cls, mode, v):
        if v['a_s'] == 0b00:
            return None
        elif v['a_s'] == 0b01:
            if v['sreg'] in [3]:
                return None
            else:
                return 16
        elif v['a_s'] == 0b10:
            return None
        elif v['a_s'] == 0b11:
            """
            if v['sreg'] in [2, 3]:
                return None
            else:
                return 16
            """
            if v['sreg'] in [0]:
                return 16
            else:
                return None
        else:
            raise NotImplementedError("unknown value v[a_s] = %d" % v['a_s'])

    def encode(self):
        return super(bs_cond_off_s, self).encode()

    def decode(self, v):
        if self.l == 0:
            self.value = None
        self.value = v
        return True


class bs_cond_off_d(bs_cond_off_s):

    @classmethod
    def flen(cls, mode, v):
        if v['a_d'] == 0:
            return None
        elif v['a_d'] == 1:
            return 16
        else:
            raise NotImplementedError("unknown value v[a_d] = %d" % v['a_d'])


class msp430_offs(imm_noarg, msp430_arg):
    parser = base_expr

    def int2expr(self, v):
        if v & ~self.intmask != 0:
            return None
        return ExprInt(v, 16)

    def decodeval(self, v):
        v <<= 1
        v += self.parent.l
        return v

    def encodeval(self, v):
        plen = self.parent.l + self.l
        assert(plen % 8 == 0)
        v -= plen // 8
        if v % 2 != 0:
            return False
        return v >> 1

    def decode(self, v):
        v = v & self.lmask
        if (1 << (self.l - 1)) & v:
            v |= ~0 ^ self.lmask
        v = self.decodeval(v)
        self.expr = ExprInt(v, 16)
        return True

    def encode(self):
        if not isinstance(self.expr, ExprInt):
            return False
        v = int(self.expr)
        if (1 << (self.l - 1)) & v:
            v = -((0xffff ^ v) + 1)
        v = self.encodeval(v)
        self.value = (v & 0xffff) & self.lmask
        return True


off_s = bs(l=16, order=-10, cls=(bs_cond_off_s,), fname = "off_s")
off_d = bs(l=16, order=-10, cls=(bs_cond_off_d,), fname = "off_d")

a_s = bs(l=2, order=-4, fname='a_s')
a_d = bs(l=1, order=-6, fname='a_d')

a_d2 = bs(l=2, order=-2, fname='a_d')

sreg = bs(l=4, order=-3, cls=(msp430_sreg_arg,), fname='sreg')
dreg = bs(l=4, order=-5, cls=(msp430_dreg_arg,), fname='dreg')

bw = bw_mn(l=1, order=-10, mn_mod=['.w', '.b'], fname='size')

bs_f1 = bs_name(
    l=4, name={
        'mov': 4, 'add': 5, 'addc': 6, 'subc': 7, 'sub': 8, 'cmp': 9,
        'dadd': 10, 'bit': 11, 'bic': 12, 'bis': 13, 'xor': 14, 'and': 15})
addop("f1", [bs_f1, sreg, a_d, bw, a_s, dreg, off_s, off_d])

bs_f2 = bs_name(l=3, name={'rrc': 0, 'rra': 2,
                           'push': 4})
addop("f2_1", [bs('000100'), bs_f2, bw, a_s, sreg, off_s])


bs_f2_nobw = bs_name(l=3, name={'swpb': 1, 'sxt': 3,
                                'call': 5})
addop("f2_2", [bs('000100'), bs_f2_nobw, bs('0'), a_s, sreg, off_s])

# Offset must be decoded in last position to have final instruction len
offimm = bs(l=10, cls=(msp430_offs,), fname="offs", order=-1)

bs_f2_jcc = bs_name(l=3, name={'jnz': 0, 'jz': 1, 'jnc': 2, 'jc': 3, 'jn': 4,
                               'jge': 5, 'jl': 6, 'jmp': 7})
addop("f2_3", [bs('001'), bs_f2_jcc, offimm])




from miasm.core.utils import size2mask
from miasm.expression.expression import ExprInt, ExprCond, ExprCompose, \
    TOK_EQUAL


def simp_ext(_, expr):
    if expr.op.startswith('zeroExt_'):
        arg = expr.args[0]
        if expr.size == arg.size:
            return arg
        return ExprCompose(arg, ExprInt(0, expr.size - arg.size))

    if expr.op.startswith("signExt_"):
        arg = expr.args[0]
        add_size = expr.size - arg.size
        new_expr = ExprCompose(
            arg,
            ExprCond(
                arg.msb(),
                ExprInt(size2mask(add_size), add_size),
                ExprInt(0, add_size)
            )
        )
        return new_expr
    return expr


def simp_flags(_, expr):
    args = expr.args

    if expr.is_op("FLAG_EQ"):
        return ExprCond(args[0], ExprInt(0, 1), ExprInt(1, 1))

    elif expr.is_op("FLAG_EQ_AND"):
        op1, op2 = args
        return ExprCond(op1 & op2, ExprInt(0, 1), ExprInt(1, 1))

    elif expr.is_op("FLAG_SIGN_SUB"):
        return (args[0] - args[1]).msb()

    elif expr.is_op("FLAG_EQ_CMP"):
        return ExprCond(
            args[0] - args[1],
            ExprInt(0, 1),
            ExprInt(1, 1),
        )

    elif expr.is_op("FLAG_ADD_CF"):
        op1, op2 = args
        res = op1 + op2
        return (((op1 ^ op2) ^ res) ^ ((op1 ^ res) & (~(op1 ^ op2)))).msb()

    elif expr.is_op("FLAG_SUB_CF"):
        op1, op2 = args
        res = op1 - op2
        return (((op1 ^ op2) ^ res) ^ ((op1 ^ res) & (op1 ^ op2))).msb()

    elif expr.is_op("FLAG_ADD_OF"):
        op1, op2 = args
        res = op1 + op2
        return (((op1 ^ res) & (~(op1 ^ op2)))).msb()

    elif expr.is_op("FLAG_SUB_OF"):
        op1, op2 = args
        res = op1 - op2
        return (((op1 ^ res) & (op1 ^ op2))).msb()

    elif expr.is_op("FLAG_EQ_ADDWC"):
        op1, op2, op3 = args
        return ExprCond(
            op1 + op2 + op3.zeroExtend(op1.size),
            ExprInt(0, 1),
            ExprInt(1, 1),
        )

    elif expr.is_op("FLAG_ADDWC_OF"):
        op1, op2, op3 = args
        res = op1 + op2 + op3.zeroExtend(op1.size)
        return (((op1 ^ res) & (~(op1 ^ op2)))).msb()

    elif expr.is_op("FLAG_SUBWC_OF"):
        op1, op2, op3 = args
        res = op1 - (op2 + op3.zeroExtend(op1.size))
        return (((op1 ^ res) & (op1 ^ op2))).msb()

    elif expr.is_op("FLAG_ADDWC_CF"):
        op1, op2, op3 = args
        res = op1 + op2 + op3.zeroExtend(op1.size)
        return (((op1 ^ op2) ^ res) ^ ((op1 ^ res) & (~(op1 ^ op2)))).msb()

    elif expr.is_op("FLAG_SUBWC_CF"):
        op1, op2, op3 = args
        res = op1 - (op2 + op3.zeroExtend(op1.size))
        return (((op1 ^ op2) ^ res) ^ ((op1 ^ res) & (op1 ^ op2))).msb()

    elif expr.is_op("FLAG_SIGN_ADDWC"):
        op1, op2, op3 = args
        return (op1 + op2 + op3.zeroExtend(op1.size)).msb()

    elif expr.is_op("FLAG_SIGN_SUBWC"):
        op1, op2, op3 = args
        return (op1 - (op2 + op3.zeroExtend(op1.size))).msb()


    elif expr.is_op("FLAG_EQ_SUBWC"):
        op1, op2, op3 = args
        res = op1 - (op2 + op3.zeroExtend(op1.size))
        return ExprCond(res, ExprInt(0, 1), ExprInt(1, 1))

    elif expr.is_op("CC_U<="):
        op_cf, op_zf = args
        return op_cf | op_zf

    elif expr.is_op("CC_U>="):
        op_cf, = args
        return ~op_cf

    elif expr.is_op("CC_S<"):
        op_nf, op_of = args
        return op_nf ^ op_of

    elif expr.is_op("CC_S>"):
        op_nf, op_of, op_zf = args
        return ~(op_zf | (op_nf ^ op_of))

    elif expr.is_op("CC_S<="):
        op_nf, op_of, op_zf = args
        return op_zf | (op_nf ^ op_of)

    elif expr.is_op("CC_S>="):
        op_nf, op_of = args
        return ~(op_nf ^ op_of)

    elif expr.is_op("CC_U>"):
        op_cf, op_zf = args
        return ~(op_cf | op_zf)

    elif expr.is_op("CC_U<"):
        op_cf, = args
        return op_cf

    elif expr.is_op("CC_NEG"):
        op_nf, = args
        return op_nf

    elif expr.is_op("CC_EQ"):
        op_zf, = args
        return op_zf

    elif expr.is_op("CC_NE"):
        op_zf, = args
        return ~op_zf

    elif expr.is_op("CC_POS"):
        op_nf, = args
        return ~op_nf

    return expr




"""
Apply simplification passes to an IR cfg
"""

import logging
import warnings
from functools import wraps
from miasm.analysis.ssa import SSADiGraph
from miasm.analysis.outofssa import UnSSADiGraph
from miasm.analysis.data_flow import DiGraphLivenessSSA
from miasm.expression.simplifications import expr_simp
from miasm.ir.ir import AssignBlock, IRBlock
from miasm.analysis.data_flow import DeadRemoval, \
    merge_blocks, remove_empty_assignblks, \
    del_unused_edges, \
    PropagateExpressions, DelDummyPhi


log = logging.getLogger("simplifier")
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter("[%(levelname)-8s]: %(message)s"))
log.addHandler(console_handler)
log.setLevel(logging.WARNING)


def fix_point(func):
    @wraps(func)
    def ret_func(self, ircfg, head):
        log.debug('[%s]: start', func.__name__)
        has_been_modified = False
        modified = True
        while modified:
            modified = func(self, ircfg, head)
            has_been_modified |= modified
        log.debug(
            '[%s]: stop %r',
            func.__name__,
            has_been_modified
        )
        return has_been_modified
    return ret_func


class IRCFGSimplifier(object):
    """
    Simplify an IRCFG
    This class applies passes until reaching a fix point
    """

    def __init__(self, lifter):
        self.lifter = lifter
        self.init_passes()

    @property
    def ir_arch(self):
        warnings.warn('DEPRECATION WARNING: use ".lifter" instead of ".ir_arch"')
        return self.lifter

    def init_passes(self):
        """
        Init the array of simplification passes
        """
        self.passes = []

    @fix_point
    def simplify(self, ircfg, head):
        """
        Apply passes until reaching a fix point
        Return True if the graph has been modified

        @ircfg: IRCFG instance to simplify
        @head: Location instance of the ircfg head
        """
        modified = False
        for simplify_pass in self.passes:
            modified |= simplify_pass(ircfg, head)
        return modified

    def __call__(self, ircfg, head):
        return self.simplify(ircfg, head)


class IRCFGSimplifierCommon(IRCFGSimplifier):
    """
    Simplify an IRCFG
    This class applies following passes until reaching a fix point:
    - simplify_ircfg
    - do_dead_simp_ircfg
    """
    def __init__(self, lifter, expr_simp=expr_simp):
        self.expr_simp = expr_simp
        super(IRCFGSimplifierCommon, self).__init__(lifter)
        self.deadremoval = DeadRemoval(self.lifter)

    def init_passes(self):
        self.passes = [
            self.simplify_ircfg,
            self.do_dead_simp_ircfg,
        ]

    @fix_point
    def simplify_ircfg(self, ircfg, _head):
        """
        Apply self.expr_simp on the @ircfg until reaching fix point
        Return True if the graph has been modified

        @ircfg: IRCFG instance to simplify
        """
        modified = ircfg.simplify(self.expr_simp)
        return modified

    @fix_point
    def do_dead_simp_ircfg(self, ircfg, head):
        """
        Apply:
        - dead_simp
        - remove_empty_assignblks
        - merge_blocks
        on the @ircfg until reaching fix point
        Return True if the graph has been modified

        @ircfg: IRCFG instance to simplify
        @head: Location instance of the ircfg head
        """
        modified = self.deadremoval(ircfg)
        modified |= remove_empty_assignblks(ircfg)
        modified |= merge_blocks(ircfg, set([head]))
        return modified


class IRCFGSimplifierSSA(IRCFGSimplifierCommon):
    """
    Simplify an IRCFG.
    The IRCF is first transformed in SSA, then apply transformations passes
    and apply out-of-ssa. Final passes of IRcfgSimplifier are applied

    This class apply following pass until reaching a fix point:
    - do_propagate_expressions
    - do_dead_simp_ssa
    """

    def __init__(self, lifter, expr_simp=expr_simp):
        super(IRCFGSimplifierSSA, self).__init__(lifter, expr_simp)

        self.lifter.ssa_var = {}
        self.all_ssa_vars = {}

        self.ssa_forbidden_regs = self.get_forbidden_regs()

        self.propag_expressions = PropagateExpressions()
        self.del_dummy_phi = DelDummyPhi()

        self.deadremoval = DeadRemoval(self.lifter, self.all_ssa_vars)

    def get_forbidden_regs(self):
        """
        Return a set of immutable register during SSA transformation
        """
        regs = set(
            [
                self.lifter.pc,
                self.lifter.IRDst,
                self.lifter.arch.regs.exception_flags
            ]
        )
        return regs

    def init_passes(self):
        """
        Init the array of simplification passes
        """
        self.passes = [
            self.simplify_ssa,
            self.do_propagate_expressions,
            self.do_del_dummy_phi,
            self.do_dead_simp_ssa,
            self.do_remove_empty_assignblks,
            self.do_del_unused_edges,
            self.do_merge_blocks,
        ]



    def ircfg_to_ssa(self, ircfg, head):
        """
        Apply the SSA transformation to @ircfg using it's @head

        @ircfg: IRCFG instance to simplify
        @head: Location instance of the ircfg head
        """
        ssa = SSADiGraph(ircfg)
        ssa.immutable_ids.update(self.ssa_forbidden_regs)
        ssa.ssa_variable_to_expr.update(self.all_ssa_vars)
        ssa.transform(head)
        self.all_ssa_vars.update(ssa.ssa_variable_to_expr)
        self.lifter.ssa_var.update(ssa.ssa_variable_to_expr)
        return ssa

    def ssa_to_unssa(self, ssa, head):
        """
        Apply the out-of-ssa transformation to @ssa using it's @head

        @ssa: SSADiGraph instance
        @head: Location instance of the graph head
        """
        cfg_liveness = DiGraphLivenessSSA(ssa.graph)
        cfg_liveness.init_var_info(self.lifter)
        cfg_liveness.compute_liveness()

        UnSSADiGraph(ssa, head, cfg_liveness)
        return ssa.graph

    @fix_point
    def simplify_ssa(self, ssa, _head):
        """
        Apply self.expr_simp on the @ssa.graph until reaching fix point
        Return True if the graph has been modified

        @ssa: SSADiGraph instance
        """
        modified = ssa.graph.simplify(self.expr_simp)
        return modified

    @fix_point
    def do_del_unused_edges(self, ssa, head):
        """
        Del unused edges of the ssa graph
        @head: Location instance of the graph head
        """
        modified = del_unused_edges(ssa.graph, set([head]))
        return modified

    def do_propagate_expressions(self, ssa, head):
        """
        Expressions propagation through ExprId in the @ssa graph
        @head: Location instance of the graph head
        """
        modified = self.propag_expressions.propagate(ssa, head)
        return modified

    @fix_point
    def do_del_dummy_phi(self, ssa, head):
        """
        Del dummy phi
        @head: Location instance of the graph head
        """
        modified = self.del_dummy_phi.del_dummy_phi(ssa, head)
        return modified

    @fix_point
    def do_remove_empty_assignblks(self, ssa, head):
        """
        Remove empty assignblks
        @head: Location instance of the graph head
        """
        modified = remove_empty_assignblks(ssa.graph)
        return modified

    @fix_point
    def do_merge_blocks(self, ssa, head):
        """
        Merge blocks with one parent/son
        @head: Location instance of the graph head
        """
        modified = merge_blocks(ssa.graph, set([head]))
        return modified

    @fix_point
    def do_dead_simp_ssa(self, ssa, head):
        """
        Apply:
        - deadrm
        - remove_empty_assignblks
        - del_unused_edges
        - merge_blocks
        on the @ircfg until reaching fix point
        Return True if the graph has been modified

        @ircfg: IRCFG instance to simplify
        @head: Location instance of the ircfg head
        """
        modified = self.deadremoval(ssa.graph)
        return modified

    def do_simplify(self, ssa, head):
        """
        Apply passes until reaching a fix point
        Return True if the graph has been modified
        """
        return super(IRCFGSimplifierSSA, self).simplify(ssa, head)

    def do_simplify_loop(self, ssa, head):
        """
        Apply do_simplify until reaching a fix point
        SSA is updated between each do_simplify
        Return True if the graph has been modified
        """
        modified = True
        while modified:
            modified = self.do_simplify(ssa, head)
            # Update ssa structs
            ssa = self.ircfg_to_ssa(ssa.graph, head)
        return ssa

    def simplify(self, ircfg, head):
        """
        Add access to "abi out regs" in each leaf block
        Apply SSA transformation to @ircfg
        Apply passes until reaching a fix point
        Apply out-of-ssa transformation
        Apply post simplification passes

        Updated simplified IRCFG instance and return it

        @ircfg: IRCFG instance to simplify
        @head: Location instance of the ircfg head
        """

        ssa = self.ircfg_to_ssa(ircfg, head)
        ssa = self.do_simplify_loop(ssa, head)
        ircfg = self.ssa_to_unssa(ssa, head)
        ircfg_simplifier = IRCFGSimplifierCommon(self.lifter)
        ircfg_simplifier.deadremoval.add_expr_to_original_expr(self.all_ssa_vars)
        ircfg_simplifier.simplify(ircfg, head)
        return ircfg



from builtins import range
from idc import get_wide_byte, get_segm_end
from idautils import Segments
from idaapi import is_mapped

from miasm.core.utils import int_to_byte
from miasm.core.bin_stream import bin_stream_str


class bin_stream_ida(bin_stream_str):
    """
    bin_stream implementation for IDA

    Don't generate xrange using address computation:
    It can raise error on overflow 7FFFFFFF with 32 bit python
    """
    def _getbytes(self, start, l=1):
        out = []
        for ad in range(l):
            offset = ad + start + self.base_address
            if not is_mapped(offset):
                raise IOError(f"not enough bytes @ offset {offset:x}")
            out.append(int_to_byte(get_wide_byte(offset)))
        return b''.join(out)

    def readbs(self, l=1):
        if self.offset + l > self.l:
            raise IOError("not enough bytes")
        content = self.getbytes(self.offset)
        self.offset += l
        return content

    def __str__(self):
        raise NotImplementedError('Not fully functional')

    def setoffset(self, val):
        self.offset = val

    def getlen(self):
        # Lazy version
        if hasattr(self, "_getlen"):
            return self._getlen
        max_addr = get_segm_end(list(Segments())[-1]  - (self.offset - self.base_address))
        self._getlen = max_addr
        return max_addr



from builtins import range
import logging

from miasm.jitter.jitload import Jitter, named_arguments
from miasm.arch.x86.sem import Lifter_X86_16, Lifter_X86_32, Lifter_X86_64
from miasm.jitter.codegen import CGen
from miasm.ir.translators.C import TranslatorC

log = logging.getLogger('jit_x86')
hnd = logging.StreamHandler()
hnd.setFormatter(logging.Formatter("[%(levelname)-8s]: %(message)s"))
log.addHandler(hnd)
log.setLevel(logging.CRITICAL)


class x86_32_CGen(CGen):
    def __init__(self, lifter):
        self.lifter = lifter
        self.PC = self.lifter.arch.regs.RIP
        self.translator = TranslatorC(self.lifter.loc_db)
        self.init_arch_C()

    def gen_post_code(self, attrib, pc_value):
        out = []
        if attrib.log_regs:
            # Update PC for dump_gpregs
            out.append("%s = %s;" % (self.C_PC, pc_value))
            out.append('dump_gpregs_32(jitcpu->cpu);')
        return out

class x86_64_CGen(x86_32_CGen):
    def gen_post_code(self, attrib, pc_value):
        out = []
        if attrib.log_regs:
            # Update PC for dump_gpregs
            out.append("%s = %s;" % (self.C_PC, pc_value))
            out.append('dump_gpregs_64(jitcpu->cpu);')
        return out

class jitter_x86_16(Jitter):

    C_Gen = x86_32_CGen

    def __init__(self, loc_db, *args, **kwargs):
        Jitter.__init__(self, Lifter_X86_16(loc_db), *args, **kwargs)
        self.vm.set_little_endian()
        self.lifter.do_stk_segm = False
        self.orig_irbloc_fix_regs_for_mode = self.lifter.irbloc_fix_regs_for_mode
        self.lifter.irbloc_fix_regs_for_mode = self.lifterbloc_fix_regs_for_mode

    def lifterbloc_fix_regs_for_mode(self, irblock, attrib=64):
        return self.orig_irbloc_fix_regs_for_mode(irblock, 64)

    def push_uint16_t(self, value):
        self.cpu.SP -= self.lifter.sp.size // 8
        self.vm.set_u16(self.cpu.SP, value)

    def pop_uint16_t(self):
        value = self.vm.get_u16(self.cpu.SP)
        self.cpu.SP += self.lifter.sp.size // 8
        return value

    def get_stack_arg(self, index):
        return self.vm.get_u16(self.cpu.SP + 4 * index)

    def init_run(self, *args, **kwargs):
        Jitter.init_run(self, *args, **kwargs)
        self.cpu.IP = self.pc


class jitter_x86_32(Jitter):

    C_Gen = x86_32_CGen

    def __init__(self, loc_db, *args, **kwargs):
        Jitter.__init__(self, Lifter_X86_32(loc_db), *args, **kwargs)
        self.vm.set_little_endian()
        self.lifter.do_stk_segm = False

        self.orig_irbloc_fix_regs_for_mode = self.lifter.irbloc_fix_regs_for_mode
        self.lifter.irbloc_fix_regs_for_mode = self.lifterbloc_fix_regs_for_mode

    def lifterbloc_fix_regs_for_mode(self, irblock, attrib=64):
        return self.orig_irbloc_fix_regs_for_mode(irblock, 64)

    def push_uint16_t(self, value):
        self.cpu.ESP -= self.lifter.sp.size // 8
        self.vm.set_u16(self.cpu.ESP, value)

    def pop_uint16_t(self):
        value = self.vm.get_u16(self.cpu.ESP)
        self.cpu.ESP += self.lifter.sp.size // 8
        return value

    def push_uint32_t(self, value):
        self.cpu.ESP -= self.lifter.sp.size // 8
        self.vm.set_u32(self.cpu.ESP, value)

    def pop_uint32_t(self):
        value = self.vm.get_u32(self.cpu.ESP)
        self.cpu.ESP += self.lifter.sp.size // 8
        return value

    def get_stack_arg(self, index):
        return self.vm.get_u32(self.cpu.ESP + 4 * index)

    def init_run(self, *args, **kwargs):
        Jitter.init_run(self, *args, **kwargs)
        self.cpu.EIP = self.pc

    # calling conventions

    # stdcall
    @named_arguments
    def func_args_stdcall(self, n_args):
        ret_ad = self.pop_uint32_t()
        args = [self.pop_uint32_t() for _ in range(n_args)]
        return ret_ad, args

    def func_ret_stdcall(self, ret_addr, ret_value1=None, ret_value2=None):
        self.pc = self.cpu.EIP = ret_addr
        if ret_value1 is not None:
            self.cpu.EAX = ret_value1
        if ret_value2 is not None:
            self.cpu.EDX = ret_value2

    def func_prepare_stdcall(self, ret_addr, *args):
        for arg in reversed(args):
            self.push_uint32_t(arg)
        self.push_uint32_t(ret_addr)

    get_arg_n_stdcall = get_stack_arg

    # cdecl
    @named_arguments
    def func_args_cdecl(self, n_args):
        ret_ad = self.pop_uint32_t()
        args = [self.get_stack_arg(i) for i in range(n_args)]
        return ret_ad, args

    def func_ret_cdecl(self, ret_addr, ret_value1=None, ret_value2=None):
        self.pc = self.cpu.EIP = ret_addr
        if ret_value1 is not None:
            self.cpu.EAX = ret_value1
        if ret_value2 is not None:
            self.cpu.EDX = ret_value2

    get_arg_n_cdecl = get_stack_arg

    # System V
    func_args_systemv = func_args_cdecl
    func_ret_systemv = func_ret_cdecl
    func_prepare_systemv = func_prepare_stdcall
    get_arg_n_systemv = get_stack_arg


    # fastcall
    @named_arguments
    def func_args_fastcall(self, n_args):
        args_regs = ['ECX', 'EDX']
        ret_ad = self.pop_uint32_t()
        args = []
        for i in range(n_args):
            args.append(self.get_arg_n_fastcall(i))
        return ret_ad, args

    def func_prepare_fastcall(self, ret_addr, *args):
        args_regs = ['ECX', 'EDX']
        for i in range(min(len(args), len(args_regs))):
            setattr(self.cpu, args_regs[i], args[i])
        remaining_args = args[len(args_regs):]
        for arg in reversed(remaining_args):
            self.push_uint32_t(arg)
        self.push_uint32_t(ret_addr)

    def get_arg_n_fastcall(self, index):
        args_regs = ['ECX', 'EDX']
        if index < len(args_regs):
            return getattr(self.cpu, args_regs[index])
        return self.get_stack_arg(index - len(args_regs))

    def syscall_args_systemv(self, n_args):
        # Documentation: http://man7.org/linux/man-pages/man2/syscall.2.html
        # args: 
        #   i386          ebx   ecx   edx   esi   edi   ebp   -
        args = [self.cpu.EBX, self.cpu.ECX, self.cpu.EDX, self.cpu.ESI,
                self.cpu.EDI, self.cpu.EBP][:n_args]
        return args

    def syscall_ret_systemv(self, value):
        # Documentation: http://man7.org/linux/man-pages/man2/syscall.2.html
        self.cpu.EAX = value


class jitter_x86_64(Jitter):

    C_Gen = x86_64_CGen
    args_regs_systemv = ['RDI', 'RSI', 'RDX', 'RCX', 'R8', 'R9']
    args_regs_stdcall = ['RCX', 'RDX', 'R8', 'R9']

    def __init__(self, loc_db, *args, **kwargs):
        Jitter.__init__(self, Lifter_X86_64(loc_db), *args, **kwargs)
        self.vm.set_little_endian()
        self.lifter.do_stk_segm = False

        self.orig_irbloc_fix_regs_for_mode = self.lifter.irbloc_fix_regs_for_mode
        self.lifter.irbloc_fix_regs_for_mode = self.lifterbloc_fix_regs_for_mode

    def lifterbloc_fix_regs_for_mode(self, irblock, attrib=64):
        return self.orig_irbloc_fix_regs_for_mode(irblock, 64)

    def push_uint64_t(self, value):
        self.cpu.RSP -= self.lifter.sp.size // 8
        self.vm.set_u64(self.cpu.RSP, value)

    def pop_uint64_t(self):
        value = self.vm.get_u64(self.cpu.RSP)
        self.cpu.RSP += self.lifter.sp.size // 8
        return value

    def get_stack_arg(self, index):
        return self.vm.get_u64(self.cpu.RSP + 8 * index)

    def init_run(self, *args, **kwargs):
        Jitter.init_run(self, *args, **kwargs)
        self.cpu.RIP = self.pc

    # calling conventions

    # stdcall
    @named_arguments
    def func_args_stdcall(self, n_args):
        args_regs = self.args_regs_stdcall
        ret_ad = self.pop_uint64_t()
        args = []
        for i in range(min(n_args, 4)):
            args.append(self.cpu.get_gpreg()[args_regs[i]])
        for i in range(max(0, n_args - 4)):
            # Take into account the shadow registers on the stack 
            # (Microsoft 64bit stdcall ABI)
            # => Skip the first 4 stack parameters
            args.append(self.get_stack_arg(4 + i))
        return ret_ad, args

    def func_prepare_stdcall(self, ret_addr, *args):
        args_regs = self.args_regs_stdcall
        for i in range(min(len(args), len(args_regs))):
            setattr(self.cpu, args_regs[i], args[i])
        remaining_args = args[len(args_regs):]
        for arg in reversed(remaining_args):
            self.push_uint64_t(arg)
        self.push_uint64_t(ret_addr)

    def func_ret_stdcall(self, ret_addr, ret_value=None):
        self.pc = self.cpu.RIP = ret_addr
        if ret_value is not None:
            self.cpu.RAX = ret_value
        return True

    # cdecl
    func_args_cdecl = func_args_stdcall
    func_ret_cdecl = func_ret_stdcall
    func_prepare_cdecl = func_prepare_stdcall

    # System V

    def get_arg_n_systemv(self, index):
        args_regs = self.args_regs_systemv
        if index < len(args_regs):
            return getattr(self.cpu, args_regs[index])
        return self.get_stack_arg(index - len(args_regs))

    @named_arguments
    def func_args_systemv(self, n_args):
        ret_ad = self.pop_uint64_t()
        args = [self.get_arg_n_systemv(index) for index in range(n_args)]
        return ret_ad, args

    func_ret_systemv = func_ret_cdecl

    def func_prepare_systemv(self, ret_addr, *args):
        args_regs = self.args_regs_systemv
        self.push_uint64_t(ret_addr)
        for i in range(min(len(args), len(args_regs))):
            setattr(self.cpu, args_regs[i], args[i])
        remaining_args = args[len(args_regs):]
        for arg in reversed(remaining_args):
            self.push_uint64_t(arg)

    def syscall_args_systemv(self, n_args):
        args = [self.cpu.RDI, self.cpu.RSI, self.cpu.RDX, self.cpu.R10,
                self.cpu.R8, self.cpu.R9][:n_args]
        return args

    def syscall_ret_systemv(self, value):
        self.cpu.RAX = value



import os
from pdb import pm
from miasm.analysis.sandbox import Sandbox_Linux_ppc32b
from miasm.core.locationdb import LocationDB
from miasm.jitter.csts import *
from miasm.jitter.jitload import log_func
import logging

# Insert here user defined methods

# Parse arguments
parser = Sandbox_Linux_ppc32b.parser(description="ELF sandboxer")
parser.add_argument("filename", help="ELF Filename")
options = parser.parse_args()

# Create sandbox
loc_db = LocationDB()
sb = Sandbox_Linux_ppc32b(loc_db, options.filename, options, globals())
log_func.setLevel(logging.ERROR)

sb.run()



import warnings
from builtins import int as int_types

from functools import reduce
from future.utils import viewitems, viewvalues

from miasm.core.utils import printable
from miasm.expression.expression import LocKey, ExprLoc


class LocationDB(object):
    """
    LocationDB is a "database" of information associated to location.

    An entry in a LocationDB is uniquely identified with a LocKey.
    Additional information which can be associated with a LocKey are:
    - an offset (uniq per LocationDB)
    - several names (each are uniqs per LocationDB)

    As a schema:
    loc_key  1 <-> 0..1  offset
             1 <-> 0..n  name

    >>> loc_db = LocationDB()
    # Add a location with no additional information
    >>> loc_key1 = loc_db.add_location()
    # Add a location with an offset
    >>> loc_key2 = loc_db.add_location(offset=0x1234)
    # Add a location with several names
    >>> loc_key3 = loc_db.add_location(name="first_name")
    >>> loc_db.add_location_name(loc_key3, "second_name")
    # Associate an offset to an existing location
    >>> loc_db.set_location_offset(loc_key3, 0x5678)
    # Remove a name from an existing location
    >>> loc_db.remove_location_name(loc_key3, "second_name")

    # Get back offset
    >>> loc_db.get_location_offset(loc_key1)
    None
    >>> loc_db.get_location_offset(loc_key2)
    0x1234
    >>> loc_db.get_location_offset("first_name")
    0x5678

    # Display a location
    >>> loc_db.pretty_str(loc_key1)
    loc_key_1
    >>> loc_db.pretty_str(loc_key2)
    loc_1234
    >>> loc_db.pretty_str(loc_key3)
    first_name
    """

    def __init__(self):
        # Known LocKeys
        self._loc_keys = set()

        # Association tables
        self._loc_key_to_offset = {}
        self._loc_key_to_names = {}
        self._name_to_loc_key = {}
        self._offset_to_loc_key = {}

        # Counter for new LocKey generation
        self._loc_key_num = 0

    def get_location_offset(self, loc_key):
        """
        Return the offset of @loc_key if any, None otherwise.
        @loc_key: LocKey instance
        """
        assert isinstance(loc_key, LocKey)
        return self._loc_key_to_offset.get(loc_key)

    def get_location_names(self, loc_key):
        """
        Return the frozenset of names associated to @loc_key
        @loc_key: LocKey instance
        """
        assert isinstance(loc_key, LocKey)
        return frozenset(self._loc_key_to_names.get(loc_key, set()))

    def get_name_location(self, name):
        """
        Return the LocKey of @name if any, None otherwise.
        @name: target name
        """
        assert isinstance(name, str)
        return self._name_to_loc_key.get(name)

    def get_or_create_name_location(self, name):
        """
        Return the LocKey of @name if any, create one otherwise.
        @name: target name
        """
        assert isinstance(name, str)
        loc_key = self._name_to_loc_key.get(name)
        if loc_key is not None:
            return loc_key
        return self.add_location(name=name)

    def get_offset_location(self, offset):
        """
        Return the LocKey of @offset if any, None otherwise.
        @offset: target offset
        """
        return self._offset_to_loc_key.get(offset)

    def get_or_create_offset_location(self, offset):
        """
        Return the LocKey of @offset if any, create one otherwise.
        @offset: target offset
        """
        loc_key = self._offset_to_loc_key.get(offset)
        if loc_key is not None:
            return loc_key
        return self.add_location(offset=offset)

    def get_name_offset(self, name):
        """
        Return the offset of @name if any, None otherwise.
        @name: target name
        """
        assert isinstance(name, str)
        loc_key = self.get_name_location(name)
        if loc_key is None:
            return None
        return self.get_location_offset(loc_key)

    def add_location_name(self, loc_key, name):
        """Associate a name @name to a given @loc_key
        @name: str instance
        @loc_key: LocKey instance
        """
        assert isinstance(name, str)
        assert loc_key in self._loc_keys
        already_existing_loc = self._name_to_loc_key.get(name)
        if already_existing_loc is not None and already_existing_loc != loc_key:
            raise KeyError("%r is already associated to a different loc_key "
                           "(%r)" % (name, already_existing_loc))
        self._loc_key_to_names.setdefault(loc_key, set()).add(name)
        self._name_to_loc_key[name] = loc_key

    def remove_location_name(self, loc_key, name):
        """Disassociate a name @name from a given @loc_key
        Fail if @name is not already associated to @loc_key
        @name: str instance
        @loc_key: LocKey instance
        """
        assert loc_key in self._loc_keys
        assert isinstance(name, str)
        already_existing_loc = self._name_to_loc_key.get(name)
        if already_existing_loc is None:
            raise KeyError("%r is not already associated" % name)
        if already_existing_loc != loc_key:
            raise KeyError("%r is already associated to a different loc_key "
                           "(%r)" % (name, already_existing_loc))
        del self._name_to_loc_key[name]
        self._loc_key_to_names[loc_key].remove(name)

    def set_location_offset(self, loc_key, offset, force=False):
        """Associate the offset @offset to an LocKey @loc_key

        If @force is set, override silently. Otherwise, if an offset is already
        associated to @loc_key, an error will be raised
        """
        assert loc_key in self._loc_keys
        already_existing_loc = self.get_offset_location(offset)
        if already_existing_loc is not None and already_existing_loc != loc_key:
            raise KeyError("%r is already associated to a different loc_key "
                           "(%r)" % (offset, already_existing_loc))
        already_existing_off = self._loc_key_to_offset.get(loc_key)
        if (already_existing_off is not None and
            already_existing_off != offset):
            if not force:
                raise ValueError(
                    "%r already has an offset (0x%x). Use 'force=True'"
                    " for silent overriding" % (
                        loc_key, already_existing_off
                    ))
            else:
                self.unset_location_offset(loc_key)
        self._offset_to_loc_key[offset] = loc_key
        self._loc_key_to_offset[loc_key] = offset

    def unset_location_offset(self, loc_key):
        """Disassociate LocKey @loc_key's offset

        Fail if there is already no offset associate with it
        @loc_key: LocKey
        """
        assert loc_key in self._loc_keys
        already_existing_off = self._loc_key_to_offset.get(loc_key)
        if already_existing_off is None:
            raise ValueError("%r already has no offset" % (loc_key))
        del self._offset_to_loc_key[already_existing_off]
        del self._loc_key_to_offset[loc_key]

    def consistency_check(self):
        """Ensure internal structures are consistent with each others"""
        assert set(self._loc_key_to_names).issubset(self._loc_keys)
        assert set(self._loc_key_to_offset).issubset(self._loc_keys)
        assert self._loc_key_to_offset == {v: k for k, v in viewitems(self._offset_to_loc_key)}
        assert reduce(
            lambda x, y:x.union(y),
            viewvalues(self._loc_key_to_names),
            set(),
        ) == set(self._name_to_loc_key)
        for name, loc_key in viewitems(self._name_to_loc_key):
            assert name in self._loc_key_to_names[loc_key]

    def find_free_name(self, name):
        """
        If @name is not known in DB, return it
        Else append an index to it corresponding to the next unknown name

        @name: string
        """
        assert isinstance(name, str)
        if self.get_name_location(name) is None:
            return name
        i = 0
        while True:
            new_name = "%s_%d" % (name, i)
            if self.get_name_location(new_name) is None:
                return new_name
            i += 1

    def add_location(self, name=None, offset=None, strict=True):
        """Add a new location in the locationDB. Returns the corresponding LocKey.
        If @name is set, also associate a name to this new location.
        If @offset is set, also associate an offset to this new location.

        Strict mode (set by @strict, default):
          If a location with @offset or @name already exists, an error will be
        raised.
        Otherwise:
          If a location with @offset or @name already exists, the corresponding
        LocKey may be updated and will be returned.
        """

        # Deprecation handling
        if isinstance(name, int_types):
            assert offset is None or offset == name
            warnings.warn("Deprecated API: use 'add_location(offset=)' instead."
                          " An additional 'name=' can be provided to also "
                          "associate a name (there is no more default name)")
            offset = name
            name = None

        # Argument cleaning
        offset_loc_key = None
        if offset is not None:
            offset = int(offset)
            offset_loc_key = self.get_offset_location(offset)

        # Test for collisions
        name_loc_key = None
        if name is not None:
            assert isinstance(name, str)
            name_loc_key = self.get_name_location(name)

        if strict:
            if name_loc_key is not None:
                raise ValueError("An entry for %r already exists (%r), and "
                                 "strict mode is enabled" % (
                                     name, name_loc_key
                                 ))
            if offset_loc_key is not None:
                raise ValueError("An entry for 0x%x already exists (%r), and "
                                 "strict mode is enabled" % (
                                     offset, offset_loc_key
                                 ))
        else:
            # Non-strict mode
            if name_loc_key is not None:
                known_offset = self.get_offset_location(name_loc_key)
                if known_offset is None:
                    if offset is not None:
                        self.set_location_offset(name_loc_key, offset)
                elif known_offset != offset:
                    raise ValueError(
                        "Location with name '%s' already have an offset: 0x%x "
                        "(!= 0x%x)" % (name, offset, known_offset)
                        )
                # Name already known, same offset -> nothing to do
                return name_loc_key

            elif offset_loc_key is not None:
                if name is not None:
                    # Check for already known name are checked above
                    return self.add_location_name(offset_loc_key, name)
                # Offset already known, no name specified
                return offset_loc_key

        # No collision, this is a brand new location
        loc_key = LocKey(self._loc_key_num)
        self._loc_key_num += 1
        self._loc_keys.add(loc_key)

        if offset is not None:
            assert offset not in self._offset_to_loc_key
            self._offset_to_loc_key[offset] = loc_key
            self._loc_key_to_offset[loc_key] = offset

        if name is not None:
            self._name_to_loc_key[name] = loc_key
            self._loc_key_to_names[loc_key] = set([name])

        return loc_key

    def remove_location(self, loc_key):
        """
        Delete the location corresponding to @loc_key
        @loc_key: LocKey instance
        """
        assert isinstance(loc_key, LocKey)
        if loc_key not in self._loc_keys:
            raise KeyError("Unknown loc_key %r" % loc_key)
        names = self._loc_key_to_names.pop(loc_key, [])
        for name in names:
            del self._name_to_loc_key[name]
        offset = self._loc_key_to_offset.pop(loc_key, None)
        self._offset_to_loc_key.pop(offset, None)
        self._loc_keys.remove(loc_key)

    def pretty_str(self, loc_key):
        """Return a human readable version of @loc_key, according to information
        available in this LocationDB instance"""
        names = self.get_location_names(loc_key)
        new_names = set()
        for name in names:
            try:
                name = name.decode()
            except AttributeError:
                pass
            new_names.add(name)
        names = new_names
        if names:
            return ",".join(names)
        offset = self.get_location_offset(loc_key)
        if offset is not None:
            return "loc_%x" % offset
        return str(loc_key)

    @property
    def loc_keys(self):
        """Return all loc_keys"""
        return self._loc_keys

    @property
    def names(self):
        """Return all known names"""
        return list(self._name_to_loc_key)

    @property
    def offsets(self):
        """Return all known offsets"""
        return list(self._offset_to_loc_key)

    def __str__(self):
        out = []
        for loc_key in self._loc_keys:
            names = self.get_location_names(loc_key)
            offset = self.get_location_offset(loc_key)
            out.append(
                "%s: %s - %s" % (
                    loc_key,
                    "0x%x" % offset if offset is not None else None,
                    ",".join(printable(name) for name in names)
                )
            )
        return "\n".join(out)

    def merge(self, location_db):
        """Merge with another LocationDB @location_db

        WARNING: old reference to @location_db information (such as LocKeys)
        must be retrieved from the updated version of this instance. The
        dedicated "get_*" APIs may be used for this task
        """
        # A simple merge is not doable here, because LocKey will certainly
        # collides

        for foreign_loc_key in location_db.loc_keys:
            foreign_names = location_db.get_location_names(foreign_loc_key)
            foreign_offset = location_db.get_location_offset(foreign_loc_key)
            if foreign_names:
                init_name = list(foreign_names)[0]
            else:
                init_name = None
            loc_key = self.add_location(offset=foreign_offset, name=init_name,
                                        strict=False)
            cur_names = self.get_location_names(loc_key)
            for name in foreign_names:
                if name not in cur_names and name != init_name:
                    self.add_location_name(loc_key, name=name)

    def canonize_to_exprloc(self, expr):
        """
        If expr is ExprInt, return ExprLoc with corresponding loc_key
        Else, return expr

        @expr: Expr instance
        """
        if expr.is_int():
            loc_key = self.get_or_create_offset_location(int(expr))
            ret = ExprLoc(loc_key, expr.size)
            return ret
        return expr

    # Deprecated APIs
    @property
    def items(self):
        """Return all loc_keys"""
        warnings.warn('DEPRECATION WARNING: use "loc_keys" instead of "items"')
        return list(self._loc_keys)

    def __getitem__(self, item):
        warnings.warn('DEPRECATION WARNING: use "get_name_location" or '
                      '"get_offset_location"')
        if item in self._name_to_loc_key:
            return self._name_to_loc_key[item]
        if item in self._offset_to_loc_key:
            return self._offset_to_loc_key[item]
        raise KeyError('unknown symbol %r' % item)

    def __contains__(self, item):
        warnings.warn('DEPRECATION WARNING: use "get_name_location" or '
                      '"get_offset_location", or ".offsets" or ".names"')
        return item in self._name_to_loc_key or item in self._offset_to_loc_key

    def loc_key_to_name(self, loc_key):
        """[DEPRECATED API], see 'get_location_names'"""
        warnings.warn("Deprecated API: use 'get_location_names'")
        return sorted(self.get_location_names(loc_key))[0]

    def loc_key_to_offset(self, loc_key):
        """[DEPRECATED API], see 'get_location_offset'"""
        warnings.warn("Deprecated API: use 'get_location_offset'")
        return self.get_location_offset(loc_key)

    def remove_loc_key(self, loc_key):
        """[DEPRECATED API], see 'remove_location'"""
        warnings.warn("Deprecated API: use 'remove_location'")
        self.remove_location(loc_key)

    def del_loc_key_offset(self, loc_key):
        """[DEPRECATED API], see 'unset_location_offset'"""
        warnings.warn("Deprecated API: use 'unset_location_offset'")
        self.unset_location_offset(loc_key)

    def getby_offset(self, offset):
        """[DEPRECATED API], see 'get_offset_location'"""
        warnings.warn("Deprecated API: use 'get_offset_location'")
        return self.get_offset_location(offset)

    def getby_name(self, name):
        """[DEPRECATED API], see 'get_name_location'"""
        warnings.warn("Deprecated API: use 'get_name_location'")
        return self.get_name_location(name)

    def getby_offset_create(self, offset):
        """[DEPRECATED API], see 'get_or_create_offset_location'"""
        warnings.warn("Deprecated API: use 'get_or_create_offset_location'")
        return self.get_or_create_offset_location(offset)

    def getby_name_create(self, name):
        """[DEPRECATED API], see 'get_or_create_name_location'"""
        warnings.warn("Deprecated API: use 'get_or_create_name_location'")
        return self.get_or_create_name_location(name)

    def rename_location(self, loc_key, newname):
        """[DEPRECATED API], see 'add_name_location' and 'remove_location_name'
        """
        warnings.warn("Deprecated API: use 'add_location_name' and "
                      "'remove_location_name'")
        for name in self.get_location_names(loc_key):
            self.remove_location_name(loc_key, name)
        self.add_location_name(loc_key, name)

    def set_offset(self, loc_key, offset):
        """[DEPRECATED API], see 'set_location_offset'"""
        warnings.warn("Deprecated API: use 'set_location_offset'")
        self.set_location_offset(loc_key, offset, force=True)

    def gen_loc_key(self):
        """[DEPRECATED API], see 'add_location'"""
        warnings.warn("Deprecated API: use 'add_location'")
        return self.add_location()

    def str_loc_key(self, loc_key):
        """[DEPRECATED API], see 'pretty_str'"""
        warnings.warn("Deprecated API: use 'pretty_str'")
        return self.pretty_str(loc_key)



import sys

from argparse import ArgumentParser
from miasm.loader import pe_init


parser = ArgumentParser(description="Create a PE from a shellcode")
parser.add_argument("filename",
                    help="x86 shellcode filename")
parser.add_argument("-p", "--pename",
                    help="new PE filename (default is 'sc_pe.exe')",
                    default="sc_pe.exe")
parser.add_argument("-w", "--word-size",
                    help="word size (default is 32 bits)",
                    choices=[32, 64],
                    type=int,
                    default=32)
args = parser.parse_args()


data = open(args.filename, 'rb').read()
pe = pe_init.PE(wsize=args.word_size)
s_text = pe.SHList.add_section(name="text", addr=0x1000, data=data)
pe.Opthdr.AddressOfEntryPoint = s_text.addr
open(args.pename, 'wb').write(bytes(pe))



from __future__ import print_function
import sys
from miasm.analysis.binary import Container
from miasm.analysis.machine import Machine
from miasm.core.locationdb import LocationDB

fdesc = open(sys.argv[1], 'rb')
loc_db = LocationDB()

# The Container will provide a *bin_stream*, bytes source for the disasm engine
# It will prodive a view from a PE or an ELF.
cont = Container.from_stream(fdesc, loc_db)

# The Machine, instantiated with the detected architecture, will provide tools
# (disassembler, etc.) to work with this architecture
machine = Machine(cont.arch)

# Instantiate a disassembler engine, using the previous bin_stream and its
# associated location DB. The assembly listing will use the binary symbols
mdis = machine.dis_engine(cont.bin_stream, loc_db=cont.loc_db)

# Run a recursive traversal disassembling from the entry point
# (do not follow sub functions by default)
addr = cont.entry_point
asmcfg = mdis.dis_multiblock(addr)

# Display each basic blocks
for block in asmcfg.blocks:
    print(block)

# Output control flow graph in a dot file
open('bin_cfg.dot', 'w').write(asmcfg.dot())
